"""Collection of methods and classes for submitting plugins to a workload manager."""
from __future__ import annotations
from pathlib import Path
from typing import cast, Union, Type, List, Optional
from .core import Job
from .local import LocalJob
from .lsf import LSFJob
from .moab import MoabJob
from .oar import OARJob
from .pbs import PBSJob
from .sge import SGEJob
from .slurm import SLURMJob


def get_job_class(system: str) -> Type[Job]:
    """Get the correct job scheduler object for a given workload manager.

    Parameters
    -----------
    system: str
        Name of the scheduling system

    Returns
    -------
    evaluation_system.api.workload_manager.core:
        Job class corresponding to the scheuler system.

    """
    job_objects: dict[str, Type[Job]] = dict(
        local=LocalJob,
        lfs=LSFJob,
        moab=MoabJob,
        oar=OARJob,
        pbs=PBSJob,
        sge=SGEJob,
        slurm=SLURMJob,
    )
    try:
        return job_objects[system]
    except KeyError:
        raise NotImplementedError(
            f"{system} scheduler not implemented, "
            f"choose from {' '.join(job_objects.keys())}"
        )


def cancel_command(system: str, job_id: int) -> str:
    """Construct a cancel command.

    Parameters:
    ===========

    system:
        Name of the workload manager system (slurm, pbs, moab, etc)
    job_id:
        The job id that is canceled
    """
    job_object = get_job_class(system)
    return f"{job_object.cancel_command} {job_id}"


def schedule_job(
    system: str,
    source: Path,
    config: dict[str, Union[str, list[str]]],
    log_directory: Union[Path, str],
    delete_job_script: bool = True,
    config_file: Optional[Path] = None,
) -> tuple[int, str]:
    """Create a scheduler object from a given scheduler configuration.

    Parameters:
    ===========
    system:
        Name of the workload manager system (slurm, pbs, moab, etc)
    source:
        Path the to source script that activates freva
    config:
        Configuration to setup a job that is submitted to the workload manager

    Returns:
    ========
    Instance of a workload manager object
    """
    job_object: Type[Job] = get_job_class(system)
    source = source.expanduser().absolute()
    ncpus = int(cast(str, config.get("cpus", "4")))
    if source.exists():
        env_extra = [f"source {source}"]
    else:
        env_extra = []
    if config_file:
        env_extra.append(f"export EVALUATION_SYSTEM_CONFIG_FILE={config_file}")
    try:
        job = job_object(
            name=cast(str, config["name"]),
            memory=cast(str, config.get("memory", "128GB")),
            walltime=cast(str, config.get("walltime", "08:00:00")),
            job_cpu=ncpus,
            queue=config["queue"],
            project=config["project"],
            log_directory=log_directory,
            job_extra=config.get("extra_options", []),
            freva_args=cast(List[str], config.get("args")),
            delete_job_script=delete_job_script,
            env_extra=env_extra,
        )
    except KeyError:
        raise ValueError("Scheduler options not properly configured")
    job.start()
    job_name = job.job_name or "worker"
    job_out = Path(log_directory) / f"{job_name}-{job.job_id}.out"
    return int(job.job_id), str(job_out.absolute())



"""Submit jobs to the PBS workload manager."""
from __future__ import annotations
import logging
import math
import os
from typing import ClassVar, Optional, Union

from .core import Job

logger = logging.getLogger(__name__)


def pbs_format_bytes_ceil(n: Union[int, float]) -> str:
    """Format bytes as text.

    PBS expects KiB, MiB or Gib, but names it KB, MB, GB  KB.

    >>> pbs_format_bytes_ceil(1)
    '1B'
    >>> pbs_format_bytes_ceil(1234)
    '1234B'
    >>> pbs_format_bytes_ceil(12345678)
    '13MB'
    >>> pbs_format_bytes_ceil(1234567890)
    '1177MB'
    >>> pbs_format_bytes_ceil(15000000000)
    '14GB'
    """
    if n >= 10 * (1024**3):
        return "%dGB" % math.ceil(n / (1024**3))
    if n >= 10 * (1024**2):
        return "%dMB" % math.ceil(n / (1024**2))
    if n >= 10 * 1024:
        return "%dkB" % math.ceil(n / 1024)
    return "%dB" % n


class PBSJob(Job):
    submit_command: ClassVar[str] = "qsub"
    cancel_command: ClassVar[str] = "qdel"
    config_name: ClassVar[str] = "pbs"

    def __init__(
        self,
        scheduler: Optional[str] = None,
        name: Optional[str] = None,
        queue: Optional[str] = None,
        project: Optional[str] = None,
        resource_spec: Optional[str] = None,
        walltime: Optional[str] = "",
        job_extra: Optional[list[str]] = None,
        **base_class_kwargs,
    ):
        job_extra = job_extra or []
        super().__init__(scheduler=scheduler, name=name, **base_class_kwargs)

        # Try to find a project name from environment variable
        project = project or os.environ.get("PBS_ACCOUNT")

        header_lines = []
        # PBS header build
        if self.job_name is not None:
            header_lines.append("#PBS -N %s" % self.job_name)
        if queue is not None:
            header_lines.append("#PBS -q %s" % queue)
        if project is not None:
            header_lines.append("#PBS -A %s" % project)
        if resource_spec is None:
            # Compute default resources specifications
            resource_spec = "select=1:ncpus=%d" % self.worker_cores
            memory_string = pbs_format_bytes_ceil(self.worker_memory)
            resource_spec += ":mem=" + memory_string
            logger.info(
                "Resource specification for PBS not set, initializing it to %s"
                % resource_spec
            )
        if resource_spec is not None:
            header_lines.append("#PBS -l %s" % resource_spec)
        if walltime is not None:
            header_lines.append("#PBS -l walltime=%s" % walltime)
        if self.log_directory is not None:
            header_lines.append("#PBS -o %s/" % self.log_directory)
            header_lines.append("#PBS -j oe")
        header_lines.extend(["#PBS %s" % arg for arg in job_extra])

        # Declare class attribute that shall be overridden
        self.job_header = "\n".join(header_lines)

        logger.debug("Job script: \n %s" % self.job_script())



"""
Created on 21.04.2015

@author: sebastian.illing@met.fu-berlin.de
"""
from django.db import models
from django.contrib.auth.models import User


class UserCrawl(models.Model):
    """
    Simple model to track user solr-crawls
    """

    class Meta:
        db_table = "solr_usercrawl"
        app_label = "solr"

    STATUS = [
        ("waiting", "waiting"),
        ("crawling", "crawling"),
        ("ingesting", "ingesting"),
        ("success", "success"),
        ("failed", "failed"),
    ]

    created = models.DateTimeField(auto_now_add=True)
    status = models.CharField(max_length=10, choices=STATUS)
    user = models.ForeignKey(
        User, on_delete=models.CASCADE
    )  # Modified 10-05-2021 Mahesh
    path_to_crawl = models.CharField(max_length=1000)
    tar_file = models.CharField(max_length=255, blank=True)
    ingest_msg = models.TextField(blank=True)



from evaluation_system.api import parameters
from evaluation_system.api.plugin import PluginAbstract
from evaluation_system.api.parameters import (
    ParameterDictionary,
    Integer,
    Float,
    String,
    Directory,
)


class ResultTagTest(PluginAbstract):
    """
    This Plugin tests the functionality of the resulttags
    """

    __parameters__ = parameters.ParameterDictionary(
        parameters.File(name="input", mandatory=True)
    )
    __short_description__ = "Test tool inserting results"
    __category__ = ""
    __tags__ = ["foo"]
    __version__ = (0, 0, 2)

    def run_tool(self, config_dict=None):
        folder = config_dict.get("folder", None)
        if folder:
            output = folder
        else:
            output = {config_dict["input"]: {"caption": "Manually added result"}}
        return self.prepare_output(output)



"""
Created on 11.03.2013

@author: Sebastian Illing / estani

This package encapsulate access to a solr instance
"""

import urllib

from evaluation_system.model.solr_core import SolrCore


class SolrFindFiles(object):
    """Encapsulate access to Solr like the find files command"""

    def __init__(self, core=None, host=None, port=None, get_status=False):
        """Create the connection pointing to the proper solr url and core.
        The default values of these parameters are setup in evaluation_system.model.solr_core.SolrCore
        and read from the configuration file.

        :param core: name of the solr core that will be used.
        :param host: hostname of the machine where the solr core is to be found.
        :param port: port number of the machine where the solr core is to be found.
        :param get_status: if the core should be contacted in an attempt to get more metadata."""
        self.solr = SolrCore(core, host=host, port=port, get_status=get_status)

    def __str__(self):  # pragma: no cover
        return "<SolrFindFiles %s>" % self.solr

    def _to_solr_query(self, partial_dict):
        """Creates a Solr query assuming the default operator is "AND". See schema.xml for that."""
        params = []
        # these are special Solr keys that we might get and we assume are not meant for the search
        special_keys = "q fl fq facet.limit".split()

        for key, value in partial_dict.items():
            if key in special_keys:
                params.append((key, value))
            else:
                if key.endswith("_not_"):
                    # handle negation
                    key = "-" + key[:-5]
                if isinstance(value, list):
                    # implies an or
                    constraint = " OR ".join(["%s:%s" % (key, v) for v in value])
                else:
                    constraint = "%s:%s" % (key, value)
                params.append(
                    (
                        "fq",
                        constraint,
                    )
                )
        return urllib.parse.urlencode(params)

    def _search(
        self,
        batch_size=10000,
        latest_version=False,
        _retrieve_metadata=False,
        **partial_dict,
    ):
        """This encapsulates the Solr call to get documents and returns an iterator providing the. The special
        parameter _retrieve_metadata will affect the first value returned by the iterator.

        :param batch_size: the amount of files to be buffered from Solr.
        :param latest_version: if the search should *try* to find the latest version from all contained here. Please note
         that we don't use this anymore. Instead we have 2 cores and this is defined directly in :class:`SolrFindFiles.search`.
         It was changed because it was slow and required too much memory.
        :param _retrieve_metadata: if set to true, the first item on the iterator is a metadata one. This is used so it can be
        known beforehand how many values are going to be returned, even before getting them all. To avoid this we might
        implement a result set object. But that would break the find_files compatibility."""
        offset = partial_dict.pop("start", 0)
        # value retrieved from sys.maxint and == 2**31-1
        max_rows = partial_dict.pop("rows", 2147483647)

        if "text" in partial_dict:
            partial_dict.update({"q": partial_dict.pop("text")})
        else:
            partial_dict.update({"q": "*:*"})

        if "fl" not in partial_dict:
            partial_dict.update({"fl": "file"})

        query = self._to_solr_query(partial_dict)
        # DEPRECATED:
        # This is not used anymore, because we have 2 different cores now
        if latest_version:  # pragma: no cover
            query += "&group=true&group.field=file_no_version&group.sort=version+desc&group.ngroups=true&group.format=simple"

        while True:
            if max_rows < batch_size:
                batch_size = max_rows
            answer = self.solr.get_json(
                "select?start=%s&rows=%s&%s" % (offset, batch_size, query)
            )
            if _retrieve_metadata:
                meta = answer["response"].copy()
                del meta["docs"]
                yield meta
                _retrieve_metadata = False
            # Not used anymore (see above)
            if latest_version:  # pragma: no cover
                offset = answer["grouped"]["file_no_version"]["doclist"]["start"]
                total = answer["grouped"]["file_no_version"]["ngroups"]
                iter_answer = answer["grouped"]["file_no_version"]["doclist"]["docs"]
            else:
                offset = answer["response"]["start"]
                total = answer["response"]["numFound"]
                iter_answer = answer["response"]["docs"]

            for item in iter_answer:
                yield item["file"]

            max_rows -= total
            if total - offset <= batch_size or max_rows <= 0:
                break  # we are done
            else:
                offset += batch_size

    @staticmethod
    def search(latest_version=True, **partial_dict):
        """It mimics the same :class:`evaluation_system.model.file.DRSFile.search` behavior.
        The implementation contacts the required Solr cores instead of contacting the file system.

        :param latest_version: defines if looking for the latest version of a file only, or for any.
        :param partial_dict: the search dictionary for solr. It might also contain some special values as
         defined in :class:`SolrFindFiles._search`
        :returns: An iterator over the results."""
        # use defaults, if other required use _search in the SolrFindFiles instance
        if latest_version:
            s = SolrFindFiles(core="latest")
        else:
            s = SolrFindFiles(core="files")
        return s._search(latest_version=False, **partial_dict)

    def _facets(self, latest_version=False, facets=None, **partial_dict):
        if facets and not isinstance(facets, list):
            if "," in facets:
                # we assume multiple values here
                facets = [f.strip() for f in facets.split(",")]
            else:
                facets = [facets]

        if "text" in partial_dict:
            partial_dict.update({"q": partial_dict.pop("text")})
        else:
            partial_dict.update({"q": "*:*"})

        query = self._to_solr_query(partial_dict)

        if facets is None:
            # get all minus what we don't want
            facets = set(self.solr.get_solr_fields()) - set(
                [
                    "",
                    "_version_",
                    "file_no_version",
                    "level",
                    "timestamp",
                    "time",
                    "creation_time",
                    "source",
                    "version",
                    "file",
                    "file_name",
                ]
            )

        if facets:
            query += (
                "&facet=true&facet.sort=index&facet.mincount=1&facet.field="
                + "&facet.field=".join(facets)
            )

        if latest_version:  # pragma: no cover (see above)
            query += "&group=true&group.field=file_no_version&group.facet=true"

        answer = self.solr.get_json("select?facet=true&rows=0&%s" % query)
        # TODO: why is there a language facit in the solr serach?
        answer = answer["facet_counts"]["facet_fields"]
        try:
            del answer["language"]
        except KeyError:
            pass
        return answer

    @staticmethod
    def facets(latest_version=True, facets=None, facet_limit=-1, **partial_dict):
        # use defaults, if other required use _search in the SolrFindFiles instance
        if latest_version:
            s = SolrFindFiles(core="latest")
        else:
            s = SolrFindFiles(core="files")
        return s._facets(facets=facets, latest_version=False, **partial_dict)



"""Definitions of custom exections and warnings."""

from contextlib import contextmanager
import logging
import sys
import warnings
from typing import Any, Callable

from evaluation_system.misc import logger


def deprecated_method(klass: str, new_method: str):
    """Show a deprication warning for a depricated method."""

    def call_deprecated_method(function: Callable[[Any], Any]) -> Callable[[Any], Any]:
        def inner(*args: Any, **kwargs: Any) -> Any:
            if logger.level <= logging.DEBUG:
                raise AttributeError(
                    f"{klass} as no attribute: {function.__name__}, use "
                    f"{new_method} instead"
                )
            msg = (
                f"The `{function.__name__}` of {klass} is deprecated and "
                "might be subject to removal in the future. Please consider "
                f"renaming the `{klass}.{function.__name__}` method to "
                f"`{klass}.{new_method}`"
            )
            logger.warning(msg)
            return function(*args, **kwargs)

        return inner

    return call_deprecated_method


class ConfigurationException(Exception):
    """Mark exceptions thrown in this package"""


class PluginNotFoundError(Exception):
    """Exeption Definition for missing Plugins."""


class ValidationError(Exception):
    """Thrown if some variable contains an improper value."""


class ParameterNotFoundError(Exception):
    """Thrown if some parameter is not in the database."""


class CommandError(Exception):
    """
    Generic exception to raise and log different fatal errors.
    """

    def __init__(self, msg):
        super(CommandError).__init__(type(self))
        self.msg = f"{msg}\nUse --help for getting help"

    def __str__(self):
        return self.msg


class PluginManagerException(Exception):
    """For all problems generating while using the plugin manager."""


@contextmanager
def hide_exception():
    """
    Suppress traceback, only print the exception type and value
    """
    default_value = getattr(sys, "tracebacklimit", 1000)
    sys.tracebacklimit = 0
    yield
    sys.tracebacklimit = default_value  # revert changes



"""Module to get information on and run Freva plgins.

To make use of any of the methods a Freva plugin has to be set up. Either
by you, the Freva admins or your collegues. If you want to create a plugin
pleas refer to the :class:`evaluation_system.api.plugin` section.
"""
from __future__ import annotations
import logging
from pathlib import Path
import json
import textwrap
from typing import Any, Union, List, Optional, NamedTuple, Tuple
import time

import appdirs
import lazy_import

from evaluation_system.model import user
from evaluation_system.misc import config, utils, logger
from evaluation_system.misc.exceptions import PluginNotFoundError
from evaluation_system.model.plugins.models import ToolPullRequest
from django.contrib.auth.models import User

pm = lazy_import.lazy_module("evaluation_system.api.plugin_manager")

CACHE_FILE = Path(appdirs.user_cache_dir()) / "freva" / "plugins.json"

PluginInfo = NamedTuple(
    "PluginInfo",
    [("name", str), ("description", str), ("parameters", List[Tuple[str, str, Any]])],
)

__all__ = ["run_plugin", "list_plugins", "plugin_doc"]


def _write_plugin_cache() -> dict[str, dict[str, Any]]:
    """Write all plugins to a file."""
    out: dict[str, dict[str, Any]] = {}
    for plugin in pm.get_plugins().keys():
        pm_instance = pm.get_plugin_instance(plugin)
        out[plugin] = {}
        out[plugin]["description"] = pm_instance.__short_description__
        out[plugin]["parameters"] = []
        for key, param in pm_instance.__parameters__._params.items():
            if param.mandatory:
                desc = f"{param.help} (mandatory)"
            else:
                desc = f"{param.help} (default: {param.default})"
            out[plugin]["parameters"].append((key, desc, param.default))

    CACHE_FILE.parent.mkdir(exist_ok=True, parents=True)
    with CACHE_FILE.open("w") as f_obj:
        json.dump(out, f_obj)
    return out


def read_plugin_cache(max_mtime: int = 180) -> list[PluginInfo]:
    """Cache plugin list."""
    CACHE_FILE.parent.mkdir(exist_ok=True, parents=True)
    CACHE_FILE.touch(exist_ok=True)
    plugin_cache = []
    if time.time() - CACHE_FILE.stat().st_mtime > max_mtime:
        cache = _write_plugin_cache()
    else:
        try:
            with CACHE_FILE.open("r") as f_obj:
                cache = json.load(f_obj)
        except:
            cache = _write_plugin_cache()
    if not cache:
        cache = _write_plugin_cache()
    for plugin, plugin_data in cache.items():
        plugin_cache.append(PluginInfo(name=plugin, **plugin_data))
    return plugin_cache


def plugin_doc(tool_name: Optional[str]) -> str:
    """Display the documentation of a given plugin.

    Parameters
    ----------
    tool_name:
        The name of the tool that should be documented

    Returns
    -------
    str:
        plugin help string.

    Raises
    ------
    PluginNotFoundError:
        if the plugin name does not exist.

    Example
    -------

    .. execute_code::

        import freva
        print(freva.plugin_doc("animator"))


    """
    tool_name = (tool_name or "").lower()
    _write_plugin_cache()
    _check_if_plugin_exists(tool_name)
    return pm.get_plugin_instance(tool_name).get_help()


def list_plugins() -> list[str]:
    """Get the plugins that are available on the system.

    Returns
    --------
    list[str]:
            List of available Freva plugins

    Example
    -------

    .. execute_code::

        import freva
        print(freva.list_plugins())
    """
    _write_plugin_cache()
    return list([k.lower() for k in pm.get_plugins().keys()])


def get_tools_list() -> str:
    """Get a list of plugins with their short description.

    Returns
    -------
    str:
        String representation of all available plugins.

    :meta private:
    """
    _write_plugin_cache()
    env = utils.get_console_size()
    # we just have to show the list and stop processing
    name_width = 0
    plugins = pm.get_plugins()
    for key in plugins:
        name_width = max(name_width, len(key))
    offset = name_width + 2
    result = []
    for key, plugin in sorted(plugins.items()):
        lines = textwrap.wrap("%s" % plugin.description, env["columns"] - offset)
        if not lines:
            lines = ["No description."]
        if len(lines) > 1:
            # multi-line
            result.append(f"{plugin.name}: {lines[0]}\n{' '*offset}\n{' '*offset}")
        else:
            result.append(f"{plugin.name}: {lines[0]}")
    return "\n".join(result)


def handle_pull_request(
    tag: Optional[str], tool_name: Optional[str]
) -> tuple[int, str]:
    """:meta private:"""
    # TODO: This method should go
    if not tag:
        return 1, 'Missing required option "--tag"'
    # create new entry in
    freva_user = user.User()
    db_user = freva_user.getUserDB().getUserId(freva_user.getName())
    pull_request = ToolPullRequest.objects.create(
        user_id=db_user, tool=tool_name, tagged_version=tag, status="waiting"
    )
    print("Please wait while your pull request is processed.")
    while pull_request.status in ["waiting", "processing"]:
        time.sleep(5)
        pull_request = ToolPullRequest.objects.get(id=pull_request.id)
    if pull_request.status == "failed":
        # TODO: Better error messages, like tag not valid or other
        return 1, "The pull request failed.\nPlease contact the admins."
    else:
        return (
            0,
            (f"{tool_name} plugin is now updated in the system.\nNew version: {tag}"),
        )


def _check_if_plugin_exists(tool_name: Optional[str]) -> None:
    """Check if a given plugin name is part of the plugin stack."""
    if tool_name in pm.get_plugins():
        return
    if not tool_name:
        error = "Available tools are:\n"
        tool_list = "\n".join(list_plugins())
    else:
        tool_list = "\n".join(utils.find_similar_words(tool_name, list_plugins()))
        error = f"{tool_name} plugin not found, did you mean:\n"
    raise PluginNotFoundError(f"\n{error}{tool_list}")


def _return_value(value: int, result: Any, return_result: bool = True) -> Any:
    if return_result:
        return value, result
    return value, ""


def run_plugin(
    tool_name: str,
    *,
    save: bool = False,
    save_config: Optional[Union[str, Path]] = None,
    show_config: bool = False,
    dry_run: bool = False,
    scheduled_id: Optional[int] = None,
    repo_version: bool = False,
    unique_output: bool = False,
    batchmode: bool = False,
    pull_request: bool = False,
    caption: str = "",
    tag: Optional[str] = None,
    return_result: bool = False,
    **options: dict[str, Union[str, float, int]],
) -> tuple[int, Any]:
    """Apply an available data analysis plugin.

    Parameters
    ----------
    tool_name:
        The name of the plugin that is to be applied.
    caption:
        Set a caption for the results.
    save:
        Save the plugin configuration to default destination.
    save_config:
        Save the plugin configuration.
    show_config:
        Show the resulting configuration (implies dry-run).
    scheduled_id:
        Run a scheduled job from database
    dry_run:
        Perform no computation. Useful for development.
    batchmode:
        Create a Batch job and submit it to the scheduling system.
    unique_output:
        Append a Freva run id to every output folder
    pull_request:
        Issue a new pull request for the tool
    return_result:
        Return the plugin result, this can be useful for pipelining.
    repo_version:
        show the version number from the repository.
    tag:
       Use git commit hash to specify a specific versrion of this tool.

    Returns
    -------
    tuple:
        Return code, and the return value of the plugin


    Example
    -------

    Run a plugin in the foreground.

    .. execute_code::

        import freva
        freva.run_plugin("animator", variable="pr", project="obs*")

    Run a plugin in the background

    .. execute_code::

        import freva
        freva.run_plugin("animator",
                         variable="pr",
                         project="observations",
                         batchmode=True)

    """
    tool_name = tool_name.lower()
    _check_if_plugin_exists(tool_name)
    if save_config:
        save_config = str(Path(save_config).expanduser().absolute())
    if pull_request:
        return _return_value(*handle_pull_request(tag, tool_name))
    if repo_version:
        (repos, version) = pm.get_plugin_version(tool_name)
        return _return_value(
            0, "Repository and version of " f":{tool_name}\n{repos}\n{version}"
        )
    options_str, tool_dict = [], {}
    for k, v in options.items():
        options_str.append(f"{k}={v}")
    if scheduled_id is None:
        tool_dict = pm.parse_arguments(tool_name, options_str)
    if logger.level == logging.DEBUG:
        tool_dict["debug"] = True
    extra_scheduler_options = tool_dict.pop("extra_scheduler_options", "")
    if caption:
        caption = caption.strip()
    if save_config or save:
        save_in = pm.write_setup(tool_name, tool_dict, config_file=save_config)
        logger.info("Configuration file saved in %s", save_in)
    elif show_config:
        return _return_value(
            0,
            pm.get_plugin_instance(tool_name).get_current_config(config_dict=tool_dict),
        )
    if scheduled_id and not dry_run:
        logger.info(
            "Running %s as scheduled in history with ID %i", tool_name, scheduled_id
        )
        out = pm.run_tool(
            tool_name, scheduled_id=scheduled_id, unique_output=unique_output
        )
        return _return_value(0, out, return_result)
    extra_options: list[str] = [
        opt.strip() for opt in extra_scheduler_options.split(",") if opt.strip()
    ]
    # now run the tool
    (error, warning) = pm.get_error_warning(tool_name)
    if warning:
        logger.warning(warning)
    if error:
        logger.error(error)
    logger.debug("Running %s with configuration: %s", tool_name, tool_dict)
    if not dry_run and not error:
        # we check if the user is external and activate batchmode
        django_user = User.objects.get(username=user.User().getName())
        if django_user.groups.filter(
            name=config.get("external_group", "noexternalgroupset")
        ).exists():
            batchmode = True
        if batchmode:
            [scheduled_id, job_file] = pm.schedule_tool(
                tool_name,
                config_dict=tool_dict,
                user=user.User(),
                caption=caption,
                extra_options=extra_options,
                unique_output=unique_output,
            )
            logger.info("Scheduled job with history id: %s", scheduled_id)
            logger.info("You can view the job's status with the command squeue")
            logger.info("Your job's progress will be shown with the command")
            logger.info("tail -f %s", job_file)
            return 0, ""
        results = pm.run_tool(
            tool_name,
            config_dict=tool_dict,
            caption=caption,
            unique_output=unique_output,
        )
        # repeat the warning at the end of the run
        # for readability don't show the warning in debug mode
        if warning:
            logger.warning(warning)
    logger.debug("Arguments: %s", options)
    logger.debug("Current configuration:\n%s", json.dumps(tool_dict, indent=4))
    return _return_value(0, results, return_result)



from difflib import SequenceMatcher
import mock
import os
import sys
import shlex
from subprocess import run, PIPE


def mockenv(**envvars):
    return mock.patch.dict(os.environ, envvars)


def run_cli(cmd):
    from freva.cli import main as main_cli

    if isinstance(cmd, str):
        cmd = shlex.split(cmd)
    return main_cli(cmd)


def similar_string(a, b, thresh=0.98):
    a = a.strip().replace("\n", " ").replace("\t", " ").replace("  ", " ").strip()
    b = b.strip().replace("\n", " ").replace("\t", " ").replace("  ", " ").strip()
    print("a", a)
    print("\n")
    print("b", b)
    ratio = SequenceMatcher(None, a, b).ratio()
    print(ratio >= thresh, thresh)
    return ratio >= thresh



from __future__ import annotations
import argparse
from pathlib import Path
import sys
from typing import Any, Optional


from evaluation_system import __version__
from evaluation_system.misc.exceptions import ValidationError
import freva
from .utils import BaseParser

CLI = "CrawlDataCli"


class CrawlDataCli(BaseParser):
    """Class that constructs the Data Crawler Argument Parser."""

    desc = "Update users project data"

    def __init__(
        self,
        command: str = "freva",
        parser: Optional[argparse.ArgumentParser] = None,
    ):
        """Construct the esgf sub arg. parser."""
        subparser = parser or argparse.ArgumentParser(
            prog=f"{command}-crawl-my-data",
            description=self.desc,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        subparser.add_argument(
            "crawl_dir",
            nargs="*",
            type=Path,
            metavar="crawl_dir",
            help="The user directory(s) that needs to be crawled",
        )
        subparser.add_argument(
            "--data-type",
            "--dtype",
            default="fs",
            choices=["fs"],
            help="The data type of the data.",
        )
        subparser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.parser = subparser
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(args: argparse.Namespace, **kwargs: Any) -> None:
        """Call the crawl my data command and print the results."""
        try:
            freva.crawl_my_data(*args.crawl_dir, dtype=args.data_type)
        except (ValidationError, ValueError) as e:
            if args.debug:
                raise e
            try:
                print(f"{e.__module__}: " f"{e.__str__()}", flush=True, file=sys.stderr)
            except AttributeError:
                print(f"{e.__repr__()}", flush=True, file=sys.stderr)
            sys.exit(1)


def main(argv: Optional[list[str]] = None) -> None:
    """Wrapper for entry point script."""
    cli = CrawlDataCli("freva")
    cli.parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    args = cli.parse_args(argv or sys.argv[1:])
    try:
        cli.run_cmd(args, **cli.kwargs)
    except KeyboardInterrupt:  # pragma: no cover
        print("KeyboardInterrupt, exiting", file=sys.stderr, flush=True)
        sys.exit(130)



"""
Created on 13.05.2016

@author: Sebastian Illing
"""

import os
import datetime
from pathlib import Path
import pytest
import mock
from tempfile import NamedTemporaryFile, TemporaryDirectory
import sys

from evaluation_system.tests import similar_string


@mock.patch("os.getpid", lambda: 12345)
def test_incomplete_abstract(dummy_plugin):
    # this is an incomplete class not implementing all required fields
    from evaluation_system.api.plugin import PluginAbstract

    class Incomplete(PluginAbstract):
        pass

    with pytest.raises(TypeError):
        Incomplete()


@mock.patch("os.getpid", lambda: 12345)
def test_complete_abstract(dummy_plugin):
    """Tests the creation of a complete implementation of the Plugin Abstract class"""
    # even though py it's just a stub, it should be complete.
    from evaluation_system.tests.mocks.dummy import DummyPlugin

    assert type(dummy_plugin) == type(DummyPlugin())


@mock.patch("os.getpid", lambda: 12345)
def test_setup_configuration(dummy_plugin):
    from evaluation_system.tests.mocks.dummy import DummyUser, DummyPlugin
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        ValidationError,
        String,
    )

    with DummyUser(random_home=True) as user:
        dummy = DummyPlugin(user=user)
        dummy.__parameters__ = ParameterDictionary(String(name="a", mandatory=True))
        # the default behavior is to check for None values and fail if found
        with pytest.raises(ValidationError):
            dummy.setup_configuration()

        # it can be turned off
        res = dummy.setup_configuration(check_cfg=False)
        assert res == {"a": None}

        # check template
        res = dummy.setup_configuration(dict(num=1), check_cfg=False)
        assert 1 == res["num"]

        # check indirect resolution
        res = dummy.setup_configuration(dict(num="${a}x", a=1), check_cfg=False)
        assert "1x" == res["num"]

        # check indirect resolution can also be turned off
        res = dummy.setup_configuration(
            dict(num="${a}x", a=1), check_cfg=False, recursion=False
        )
        assert "${a}x" == res["num"]

        # check user special values work
        res = dummy.setup_configuration(dict(num="$USER_BASE_DIR"), check_cfg=False)
        assert user.getUserBaseDir() == res["num"]

        res = dummy.setup_configuration(
            dict(num="$USER_BASE_DIR"), check_cfg=False, substitute=False
        )
        assert "$USER_BASE_DIR" == res["num"]


@mock.patch("os.getpid", lambda: 12345)
def test_parse_arguments(dummy_plugin):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        String,
        Integer,
        ValidationError,
        Bool,
    )

    dummy = dummy_plugin
    dummy.__parameters__ = ParameterDictionary(String(name="a"), String(name="b"))
    res = dummy.__parameters__.parse_arguments("a=1 b=2".split())
    assert res == dict(a="1", b="2")
    dummy.__parameters__ = ParameterDictionary(
        Integer(name="a", default=0), Integer(name="b", default=0)
    )
    res = dummy.__parameters__.parse_arguments("a=1 b=2".split())
    assert res == dict(a=1, b=2)
    # even if the default value is different, the metadata can define the type
    dummy.__parameters__ = ParameterDictionary(
        Integer(name="a", default="1"), Integer(name="b", default=2)
    )
    res = dummy.__parameters__.parse_arguments("a=1 b=2".split())
    assert res == dict(a=1, b=2)
    # more arguments than those expected
    dummy.__parameters__ = ParameterDictionary(Integer(name="a", default="1"))
    with pytest.raises(ValidationError):
        dummy.__parameters__.parse_arguments("a=1 b=2".split())
    dummy.__parameters__ = ParameterDictionary(Bool(name="a"))
    for arg, parsed_val in [
        ("a=1", True),
        ("a=true", True),
        ("a=TRUE", True),
        ("a=0", False),
        ("a=false", False),
        ("a=False", False),
    ]:
        res = dummy.__parameters__.parse_arguments(arg.split())
        assert res == dict(a=parsed_val), "Error when parsing %s, got %s" % (arg, res)


@mock.patch("os.getpid", lambda: 12345)
def test_parse_metadict(dummy_plugin):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        String,
        Integer,
        Bool,
        Float,
        ValidationError,
    )

    dummy = dummy_plugin
    for d, res_d in [
        (Integer(name="a", default=0), 1),
        (Integer(name="a"), 1),
        (Integer(name="a", default="0"), 1),
        (String(name="a"), "1"),
        (String(name="a", default=1), "1"),
        (Bool(name="a"), True),
        (Float(name="a"), float("1")),
    ]:
        dummy.__parameters__ = ParameterDictionary(d)
        res = dummy.parse_config_str_value("a", "1")
        assert res == res_d

    # check errors
    # Wrong type
    dummy.__parameters__ = ParameterDictionary(Integer(name="a"))
    with pytest.raises(ValidationError):
        dummy.parse_config_str_value("a", "d")
    # wrong key
    with pytest.raises(ValidationError):
        dummy.parse_config_str_value("b", "1")


@mock.patch("os.getpid", lambda: 12345)
def test_read_config_parser(dummy_plugin):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        String,
        Integer,
        ValidationError,
    )
    from configparser import ConfigParser, ExtendedInterpolation
    from io import StringIO

    conf = ConfigParser(interpolation=ExtendedInterpolation())
    conf_str = "[DummyPlugin]\na=42\nb=text"
    conf.read_file(StringIO(conf_str))
    dummy = dummy_plugin
    # check parsing
    for d, res_d in [
        ([Integer(name="a")], dict(a=42)),
        ([String(name="a")], dict(a="42")),
        ([Integer(name="a"), String(name="b")], dict(a=42, b="text")),
    ]:
        dummy.__parameters__ = ParameterDictionary(*d)
        res = dummy.read_from_config_parser(conf)
        assert res == {**res_d, **{"extra_scheduler_options": ""}}
    # check errors
    # wrong type
    dummy.__parameters__ = ParameterDictionary(Integer(name="b"))
    with pytest.raises(ValidationError):
        dummy.read_from_config_parser(conf)


@mock.patch("os.getpid", lambda: 12345)
def test_save_config(dummy_plugin):
    from evaluation_system.api.parameters import ParameterDictionary, Integer, String
    from io import StringIO

    batchmode_options = """#: Set additional options for the job submission to the workload manager (,
#:  seperated). Note: batchmode and web only.
#extra_scheduler_options="""

    res_str = StringIO()
    dummy = dummy_plugin
    dummy.__parameters__ = ParameterDictionary(
        Integer(name="a", help=""), Integer(name="b", help="")
    )
    tests = [
        (dict(a=1), "[DummyPlugin]\na=1"),
        (dict(a="1"), "[DummyPlugin]\na=1"),
        (dict(b=2), "[DummyPlugin]\nb=2"),
    ]
    for t, res in tests:
        res_str.truncate(0)
        dummy.save_configuration(res_str, t)
        assert res_str.getvalue().strip("\x00").strip() == res

    dummy.__parameters__ = ParameterDictionary(
        Integer(name="a", help="This is \na test"),
        Integer(name="b", help="Also\na\ntest."),
    )
    res_str.truncate(0)
    dummy.save_configuration(res_str, {"a": 1})
    assert (
        res_str.getvalue().strip("\x00").strip().strip("\x00")
        == "[DummyPlugin]\n#: This is\n#: a test\na=1"
    )
    res_str.truncate(0)
    dummy.save_configuration(res_str, {"a": 1}, include_defaults=True)
    assert (
        res_str.getvalue().strip("\x00").strip()
        == """[DummyPlugin]
#: This is
#: a test
a=1

#: Also
#: a
#: test.
#b=None

"""
        + batchmode_options
    )

    dummy.__parameters__ = ParameterDictionary(
        Integer(
            name="a",
            help="""This is a very long and complex explanation so that we could test how it is transformed into a proper configuration file that:
a) is understandable
b) Retains somehow the format
c) its compact
We'll have to see how that works out...""",
        ),
        String(
            name="b",
            mandatory=True,
            help="Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet.",
        ),
        String(
            name="c",
            help=r"""This is to check the format is preserved
             \..,.-
             .\   |         __
             .|    .-     /  .;
       _      _|      \__/..     \ 
      \ ...\|   X Hamburg        :_
       |                           \ 
      /                            /
     -.                           |
      \  X Rheine      Berlin X    \ 
   __/                              |
  |                                 /
  |                                 \ 
 /                                   |
 \     X Cologne        Dresden  X . ,
  \                            ._-. .
 /                        __.-/
 |         X Frankfurt    \ 
  \                        \ 
   \                        \ 
    ...,.                    \ 
        /                     \.
       /                       ,.
      /                      ./
     |         Munich X     |
     \,......,__  __,  __.-. .
                \/   -/     ..
dj1yfk""",
        ),
    )
    res_str.truncate(0)
    dummy.save_configuration(res_str, {"a": 1}, include_defaults=True)


@mock.patch("os.getpid", lambda: 12345)
def test_read_config(dummy_plugin):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        String,
        Float,
    )

    dummy = dummy_plugin
    dummy.__parameters__ = ParameterDictionary(
        Integer(name="a"),
        String(name="b", default="test"),
        Float(name="other", default=1.4),
    )
    for resource, expected_dict in [
        ("[DummyPlugin]\na=1", dict(a=1, b="test", other=1.4)),
        ("[DummyPlugin]\na= 1 \n", dict(a=1, b="test", other=1.4)),
        ("[DummyPlugin]\na=1\nb=2", dict(a=1, b="2", other=1.4)),
        (
            "[DummyPlugin]\n#a=1\nb=2\n#asd\nother=1e10",
            dict(a=None, b="2", other=1e10),
        ),
        (
            "[DummyPlugin]\na=-2\nb=blah blah blah",
            dict(a=-2, b="blah blah blah", other=1.4),
        ),
    ]:
        with NamedTemporaryFile() as tf:
            open(tf.name, "w").write(resource)
            with open(tf.name, "r") as f:
                conf_dict = dummy.read_configuration(f)
                assert conf_dict == {**expected_dict, **{"extra_scheduler_options": ""}}


@mock.patch("os.getpid", lambda: 12345)
def testSubstitution(dummy_plugin):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        String,
        InputDirectory,
    )

    dummy = dummy_plugin
    dummy.__parameters__ = ParameterDictionary(
        Integer(name="a"),
        String(name="b", default="value:$a"),
        InputDirectory(name="c", default="$USER_OUTPUT_DIR"),
    )
    cfg_str = dummy.get_current_config({"a": 72})
    assert "value:72" in cfg_str
    assert "b: - (default: value:$a [value:72])" in cfg_str
    assert "c: - (default: $USER_OUTPUT_DIR [" in cfg_str


@mock.patch("os.getpid", lambda: 12345)
def test_help(dummy_plugin):
    from evaluation_system.api.parameters import ParameterDictionary, Integer, String

    dummy = dummy_plugin
    dummy.__version__ = (1, 2, 3)
    dummy.__parameters__ = ParameterDictionary(
        Integer(name="a", default=1, help="This is the value of a"),
        String(name="b", help="This is not the value of b"),
        String(
            name="example",
            default="test",
            help="let's hope people write some useful help...",
        ),
    )

    descriptions = [
        ("__short_description__", "A short Description"),
        ("__long_description__", "A long Description"),
    ]
    for desc in descriptions:
        setattr(dummy, desc[0], desc[1])
        resource = (
            """DummyPlugin (v1.2.3): %s
Options:
a       (default: 1)
    This is the value of a
b       (default: <undefined>)
    This is not the value of b
example (default: test)
    let's hope people write some useful help...#
extra_scheduler_options (default: )
                        Set additional options for the job submission to the
                        workload manager (, seperated). Note: batchmode and web
                        only."""
            % desc[1]
        )
        assert similar_string(dummy.get_help().strip(), resource.strip(), 0.9)


@mock.patch("os.getpid", lambda: 12345)
def test_show_config(dummy_plugin):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        String,
        Float,
    )
    from evaluation_system.tests.mocks.dummy import DummyUser, DummyPlugin

    user = DummyUser(random_home=True)
    dummy = DummyPlugin(user=user)

    dummy.__parameters__ = ParameterDictionary(
        Integer(name="a", mandatory=True, help="This is the value of a"),
        String(name="b", default="test", help="This is not the value of b"),
        Float(
            name="other",
            default=1.4,
        ),
    )
    assert similar_string(
        dummy.get_current_config(),
        """           a: - *MUST BE DEFINED!*
                      b: - (default: test)
                  other: - (default: 1.4)
extra_scheduler_options: - (default: )""",
    )
    assert similar_string(
        dummy.get_current_config(config_dict=dict(a=2123123)),
        """           a: 2123123
                      b: - (default: test)
                  other: - (default: 1.4)
extra_scheduler_options: - (default: )""",
    )
    assert similar_string(
        dummy.get_current_config(config_dict=dict(a=2123123)),
        """         a: 2123123
                   b: - (default: test)
               other: - (default: 1.4)
extra_scheduler_options: - (default: )""",
    )
    res = dummy.get_current_config(config_dict=dict(a="/tmp$USER_PLOTS_DIR"))
    cmp_str = (
        "                          a: /tmp$USER_PLOTS_DIR [/tmp"
        + user.getUserPlotsDir("DummyPlugin")
        + "]\n                     b: - (default: test)\n"
        + "                     other: - (default: 1.4)"
        + "\nextra_scheduler_options: - (default: )"
    )
    assert similar_string(cmp_str, res)
    user.cleanRandomHome()


@mock.patch("os.getpid", lambda: 12345)
def test_usage(dummy_plugin):
    from evaluation_system.api.parameters import ParameterDictionary, Integer, String

    dummy = dummy_plugin
    dummy.__parameters__ = ParameterDictionary(
        Integer(name="a", help="This is very important"),
        Integer(name="b", default=2),
        String(name="c", help="Well this is just an x..."),
    )
    def_config = dummy.setup_configuration(check_cfg=False)
    from io import StringIO

    resource = StringIO()
    dummy.save_configuration(resource, def_config)
    res_str1 = resource.getvalue()
    resource.truncate(0)
    dummy.save_configuration(resource)
    res_str2 = resource.getvalue()
    assert res_str1.strip("\x00") == res_str2.strip("\x00")


@mock.patch("os.getpid", lambda: 12345)
def test_run(dummy_plugin):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        String,
        InputDirectory,
    )
    from evaluation_system.tests.mocks.dummy import DummyPlugin

    dummy = dummy_plugin
    # no config
    dummy._run_tool()
    assert len(DummyPlugin._runs) == 1
    run = DummyPlugin._runs[0]
    assert not run
    DummyPlugin._runs = []

    # direct config
    dummy._run_tool(config_dict=dict(the_number=42))
    assert len(DummyPlugin._runs) == 1
    run = DummyPlugin._runs[0]
    assert "the_number" in run
    assert run["the_number"] == 42
    DummyPlugin._runs = []


@mock.patch("os.getpid", lambda: 12345)
def test_get_class_base_dir(dummy_plugin):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        String,
        InputDirectory,
    )

    dummy = dummy_plugin
    import evaluation_system.tests.mocks
    import os
    import re

    assert evaluation_system.tests.mocks.dummy.__file__.startswith(dummy.class_basedir)
    # module name should be getClassBaseDir() + modulename_with_"/"_instead_of_"." + ".pyc" or ".py"
    module_name = os.path.abspath(evaluation_system.tests.mocks.dummy.__file__)[
        len(dummy.class_basedir) + 1 :
    ].replace("/", ".")
    print(module_name, "blablabla")
    module_name = re.sub("\\.pyc?$", "", module_name)
    assert module_name == "evaluation_system.tests.mocks.dummy"


@mock.patch("os.getpid", lambda: 12345)
def test_special_variables():
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        String,
        InputDirectory,
    )
    from evaluation_system.tests.mocks.dummy import DummyPlugin

    dummy = DummyPlugin()
    special_vars = dict(
        sv_USER_BASE_DIR="$USER_BASE_DIR",
        sv_USER_OUTPUT_DIR="$USER_OUTPUT_DIR",
        sv_USER_PLOTS_DIR="$USER_PLOTS_DIR",
        sv_USER_CACHE_DIR="$USER_CACHE_DIR",
        sv_SYSTEM_DATE="$SYSTEM_DATE",
        sv_SYSTEM_DATETIME="$SYSTEM_DATETIME",
        sv_SYSTEM_TIMESTAMP="$SYSTEM_TIMESTAMP",
        sv_SYSTEM_RANDOM_UUID="$SYSTEM_RANDOM_UUID",
    )

    result = dict(
        [
            (k, v)
            for k, v in dummy.setup_configuration(
                config_dict=special_vars, check_cfg=False
            ).items()
            if k in special_vars
        ]
    )
    import re

    assert re.match("[0-9]{8}$", result["sv_SYSTEM_DATE"])
    assert re.match("[0-9]{8}_[0-9]{6}$", result["sv_SYSTEM_DATETIME"])
    assert re.match("[0-9]{9,}$", result["sv_SYSTEM_TIMESTAMP"])
    assert re.match(
        "[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}$",
        result["sv_SYSTEM_RANDOM_UUID"],
    )


@mock.patch("os.getpid", lambda: 12345)
def test_compose_command():

    from evaluation_system.tests.mocks.dummy import DummyPlugin

    dummy_plugin = DummyPlugin()
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        String,
        InputDirectory,
    )

    command = dummy_plugin.compose_command(
        config_dict={"the_number": 22}, caption="This is the caption"
    )
    assert similar_string(
        " ".join(command),
        (
            "dummpyplugin the_number=22 something=test other=1.4 variable=tas "
            "--caption 'This is the caption' --unique_output True"
        ),
    )


@mock.patch("os.getpid", lambda: 12345)
def test_append_unique_output():
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        String,
        InputDirectory,
    )
    from evaluation_system.misc import config
    from evaluation_system.tests.mocks.dummy import DummyPlugin

    try:
        dummy_plugin = DummyPlugin()
        config._config[config.DIRECTORY_STRUCTURE_TYPE] = "scratch"
        config_dict = {
            "the_number": 42,
            "input": "/my/input/dir",
            "directory_structure_type": "scratch",
        }
        dummy_plugin.rowid = 1
        new_config = dummy_plugin._append_unique_id(config_dict.copy(), True)
        assert new_config["input"] == "/my/input/dir"
        new_config = dummy_plugin._append_unique_id(config_dict.copy(), False)
        assert new_config["input"] == "/my/input/dir"
    finally:
        config._config[config.DIRECTORY_STRUCTURE_TYPE] = "local"
        config.reloadConfiguration()


@mock.patch("os.getpid", lambda: 12345)
def test_run_tool(dummy_plugin):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        String,
        InputDirectory,
    )

    result = dummy_plugin._run_tool({"the_answer": 42})
    assert result == {
        "/tmp/dummyfile1": dict(type="plot"),
        "/tmp/dummyfile2": dict(type="data"),
    }


@mock.patch("os.getpid", lambda: 12345)
def test_prepare_output(dummy_plugin):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        String,
        InputDirectory,
    )

    fn = Path(__file__).absolute().parent / "test_output" / "vecap_test_output.pdf"
    types_to_check = [
        {"suffix": ".jpg", "type": "plot", "todo": "copy"},
        {"suffix": ".eps", "type": "plot", "todo": "convert"},
        {"suffix": ".zip", "type": "pdf", "todo": "copy"},
        {"suffix": ".pdf", "type": "pdf", "todo": "copy", "fn": fn},
    ]
    for check in types_to_check:
        from evaluation_system.tests.mocks.result_tags import ResultTagTest

        plugin = ResultTagTest()
        if check["suffix"] == ".pdf":
            fn = check.pop("fn")
            meta_data = plugin._run_tool({"input": str(fn)})
            meta_data = meta_data[list(meta_data.keys())[0]]
        else:
            with NamedTemporaryFile(mode="wb", suffix=check["suffix"]) as fn:
                meta_data = plugin._run_tool({"input": fn.name})[fn.name]
        assert meta_data["todo"] == check["todo"]
        assert meta_data["type"] == check["type"]
        assert meta_data["caption"] == "Manually added result"
    meta_data = plugin._run_tool({"folder": str(fn.parent)})
    assert len(meta_data) == 2


@mock.patch("os.getpid", lambda: 12345)
def test_call(dummy_plugin, capsys):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        String,
        InputDirectory,
    )

    _ = dummy_plugin.call("echo /bin/bash")  # .strip('\n')
    assert "/bin/bash" in capsys.readouterr().out


@mock.patch("os.getpid", lambda: 12345)
def test_link_mydata(dummy_plugin, dummy_solr):
    with TemporaryDirectory() as td:
        for file in dummy_solr.files:
            f = Path(td) / file
            f.parent.mkdir(exist_ok=True, parents=True)
            with f.open("w") as f:
                f.write(" ")
        # dummy_plugin.linkmydata(outputdir=td)



"""
.. moduleauthor:: Sebastian Illing

This modules encapsulates all access to databases.
"""

import evaluation_system.model.history.models as hist
import evaluation_system.model.plugins.models as pin

from django.contrib.auth.models import User
from django.db import transaction

from datetime import datetime
import json
import re
import pandas as pd
import socket
from evaluation_system.misc import config
from evaluation_system.model.history.models import Configuration

from evaluation_system.misc import logger as log
import evaluation_system.settings.database

TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S.%f"


def timestamp_to_string(datetime_obj):
    return datetime_obj.strftime(TIMESTAMP_FORMAT)


def timestamp_from_string(date_string):
    """Create a datetime object from a given datetime string."""
    if date_string is None:
        return
    try:
        return pd.Timestamp(date_string).to_pydatetime()
    except:
        raise ValueError("Can't parse a date from '%s'" % date_string)


# class HistoryResultEntry(object):
#     """
#     This class encapsulates the access to the results.
#     """
#     def __init__(self, row):
#         self.id = row[0]
#         self.history_id_id = row[1]
#         self.output_file = row[2]
#         self.preview_file = row[3]
#         self.filetype = row[4]
#
#
# class HistoryTagEntry(object):
#     """
#     This class encapsulates the access to the HistoryTag entries.
#     """
#     def __init__(self, row):
#         self.id = row[0]
#         self.history_id_id = row[1]
#         self.type = row[2]
#         self.uid = row[3]
#         self.text = row[4]


class UserDB(object):
    """Encapsulates access to the local DB of a single user."""

    def __init__(self, user):
        self._user = user

    @transaction.atomic
    @transaction.atomic
    def storeHistory(
        self,
        tool,
        config_dict,
        uid,
        status,
        slurm_output=None,
        result=None,
        flag=None,
        version_details=None,
        caption=None,
    ):
        """Store a an analysis run into the DB.

        :type tool: :class:`evaluation_system.api.plugin.pluginAbstract`
        :param tool: the plugin for which we are storing the information.
        :param config_dict: dictionary with the configuration used for this run,
        :param uid: the user id (useful in a global database)
        :param status: the process status
        :param result: dictionary with the results (created files)."""
        if result is None:
            result = {}
        if slurm_output is None:
            slurm_output = 0
        if flag is None:
            flag = 0
        if version_details is None:
            version_details = 1
        toolname = tool.__class__.__name__.lower()
        version = repr(tool.__version__)
        for key in config.exclude:
            config_dict.pop(key, "")
        newentry = hist.History(
            timestamp=datetime.now(),
            tool=toolname,
            version=version,
            configuration=json.dumps(config_dict),
            slurm_output=slurm_output,
            uid_id=uid,
            status=status,
            flag=flag,
            version_details_id=version_details,
        )
        # set caption
        if caption:
            newentry.caption = caption

        tool.__parameters__.synchronize(toolname)

        try:
            newentry.save()

            for p in tool.__parameters__._params.values():
                if p.name not in config.exclude:
                    name = p.name
                    param = Configuration(
                        history_id_id=newentry.id,
                        parameter_id_id=p.id,
                        value=json.dumps(config_dict[name]),
                        is_default=p.is_default,
                    )
                    param.save()
        except Exception as e:
            raise e

        return newentry.id

    def scheduleEntry(self, row_id, uid, slurmFileName, status=None):
        """Schedule a tool for a future application
        Parameter:
        ----------
        row_id:
            The index in the history table
        uid:
            the user id
        slurmFileName:
            The std out file belonging to the history entry
        stauts:
            Overwrite the default status (scheduled) with this status.
        """

        h = hist.History.objects.get(
            id=row_id, uid_id=uid, status=hist.History.processStatus.not_scheduled
        )

        h.slurm_output = slurmFileName
        h.host = socket.gethostbyname(socket.gethostname())
        h.status = status or hist.History.processStatus.scheduled

        h.save()

    class ExceptionStatusUpgrade(Exception):
        """
        Exception class for failing status upgrades
        """

        def __init__(self, msg="Status could not be upgraded"):
            super(UserDB.ExceptionStatusUpgrade, self).__init__(msg)

    def upgradeStatus(self, row_id, uid, status):
        """
        :param row_id: The index in the history table
        :param uid: the user id
        :param status: the new status
        After validation the status will be upgraded.
        """

        h = hist.History.objects.get(pk=row_id, uid_id=uid)

        if h.status < status:
            raise self.ExceptionStatusUpgrade("Tried to downgrade a status")

        h.status = status

        h.save()

    def changeFlag(self, row_id, uid, flag):
        """
        :param row_id: The index in the history table
        :param uid: the user id
        :param flag: the new flag
        After validation the status will be upgraded.
        """

        h = hist.History.objects.get(id=row_id, uid_id=uid)

        h.flag = flag

        h.save()

    def getHistory(
        self, tool_name=None, limit=-1, since=None, until=None, entry_ids=None, uid=None
    ):
        """Returns the stored history (run analysis) for the given tool.

        :type tool_name: str
        :param tool_name: name of the tool for which the information will be gathered (if None, then everything is returned).
        :type limit: int
        :param limit: Amount of rows to be returned (if < 0, return all).
        :type since: datetime.datetime
        :param since: Return only items stored after this date
        :type until: datetime.datetime
        :param until: Return only  items stored before this date
        :param entry_ids: ([int] or int) id or list thereof to be selected
        :returns: ([:class:`HistoryEntry`]) list of entries that match the query."""
        filter_dict = {}

        if entry_ids is not None:
            if isinstance(entry_ids, int):
                entry_ids = [entry_ids]
            filter_dict["id__in"] = entry_ids

        if tool_name is not None:
            filter_dict["tool"] = tool_name

        if since is not None:
            filter_dict["timestamp__gte"] = since

        if until is not None:
            filter_dict["timestamp__lte"] = until

        if uid is not None:
            filter_dict["uid_id"] = uid

        o = hist.History.objects.filter(**filter_dict).order_by("-id")

        if limit > 0:
            o = o[:limit]

        return o

    def addHistoryTag(self, hrowid, tagType, text, uid=None):
        """
        :type hrowid: integer
        :param hrowid: the row id of the history entry where the results belong to
        :type tagType: integer
        :param tagType: the kind of tag
        :type: text: string
        :param: text: the text belonging to the tag
        :type: uid: string
        :param: uid: the user, default: None
        """

        h = hist.HistoryTag(history_id_id=hrowid, type=tagType, text=text)

        if uid is not None:
            h.uid_id = uid

        h.save()

    def updateHistoryTag(self, trowid, tagType=None, text=None, uid=None):
        """
        :type trowid: integer
        :param trowid: the row id of the tag
        :type tagType: integer
        :param tagType: the kind of tag
        :type: text: string
        :param: text: the text belonging to the tag
        :type: uid: string
        :param: uid: the user, default: None
        """

        h = hist.HistoryTag.objects.get(id=trowid, uid_id=uid)

        if tagType is not None:
            h.type = tagType

        if text is not None:
            h.text = text

        h.save()

    def storeResults(self, rowid, results):
        """
        :type rowid: integer
        :param rowid: the row id of the history entry where the results belong to
        :type results: dict with entries {str : dict}
        :param results: meta-dictionary with meta-data dictionaries assigned to the file names.
        """
        reg_ex = None

        # regex to get the relative path
        preview_path = config.get(config.PREVIEW_PATH, None)
        expression = "(%s\\/*){1}(.*)" % re.escape(preview_path)

        # only try to create previews, when a preview path is given
        if preview_path:
            reg_ex = re.compile(expression)

        for file_name in results:
            metadata = results[file_name]

            type_name = metadata.get("type", "")
            type_number = hist.Result.Filetype.unknown

            preview_path = metadata.get("preview_path", "")
            preview_file = ""

            if preview_path and reg_ex is not None:
                # We store the relative path for previews only.
                # Which allows us to move the preview files to a different folder.
                preview_file = reg_ex.match(preview_path).group(2)

            if type_name == "plot":
                type_number = hist.Result.Filetype.plot
            elif type_name == "data":
                type_number = hist.Result.Filetype.data

            h = hist.Result(
                history_id_id=rowid,
                output_file=file_name,
                preview_file=preview_file,
                file_type=type_number,
            )

            h.save()

            result_id = h.pk
            self._storeResultTags(result_id, metadata)

    def _storeResultTags(self, result_id, metadata):
        """
        :type result_id: integer
        :param result_id: the id of the result entry where the tag belongs to
        :type metadata: dict with entries {str : dict}
        :param metadata: meta-dictionary with meta-data dictionaries assigned to the file names.
        """

        data_to_store = []

        # append new tags here
        caption = metadata.get("caption", None)

        if caption:
            data_to_store.append(
                hist.ResultTag(
                    result_id_id=result_id,
                    type=hist.ResultTag.flagType.caption,
                    text=caption,
                )
            )

        hist.ResultTag.objects.bulk_create(data_to_store)

    def getVersionId(
        self,
        toolname,
        version,
        repos_api,
        internal_version_api,
        repos_tool,
        internal_version_tool,
    ):
        repository = "%s;%s" % (repos_tool, repos_api)

        retval = None

        try:
            p = pin.Version.objects.filter(
                tool=toolname,
                version=version,
                internal_version_tool=internal_version_tool[:40],
                internal_version_api=internal_version_api[:40],
                repository=repository,
            )[0]

            retval = p.pk

        except (IndexError, pin.Version.DoesNotExist) as e:
            pass

        return retval

    def newVersion(
        self,
        toolname,
        version,
        repos_api,
        internal_version_api,
        repos_tool,
        internal_version_tool,
    ):
        repository = "%s;%s" % (repos_tool, repos_api)

        p = pin.Version(
            timestamp=datetime.now(),
            tool=toolname,
            version=version,
            internal_version_tool=internal_version_tool,
            internal_version_api=internal_version_api,
            repository=repository,
        )

        p.save()

        result_id = p.pk

        return result_id

    def getUserId(self, username):
        retval = 0

        try:
            u = User.objects.get(username=username)

            retval = u.pk
        except User.DoesNotExist:
            pass

        return retval

    def updateUserLogin(self, row_id, email=None):
        u = User.objects.get(id=row_id)

        u.last_login = datetime.now()

        if email is not None:
            u.email = email

        u.save()

    def createUser(
        self,
        username,
        email="-",
        first_name="",
        last_name="",
    ):

        timestamp = datetime.now()

        u = User(
            username=username,
            password="NoPasswd",
            date_joined=timestamp,
            last_login=timestamp,
            first_name=first_name,
            last_name=last_name,
            email=email,
            is_active=1,
            is_staff=0,
            is_superuser=0,
        )

        u.save()

    def create_user_crawl(self, crawl_dir, username):
        from evaluation_system.model.solr_models.models import UserCrawl

        crawl = UserCrawl(
            status="waiting", user_id=self.getUserId(username), path_to_crawl=crawl_dir
        )
        crawl.save()
        return crawl.id



"""A Python module to access the apache solr databrowser."""
from __future__ import annotations
from pathlib import Path

from evaluation_system.misc import logger
from evaluation_system.model.solr import SolrFindFiles

from typing import Any, Optional, Union, Iterator, overload
from typing_extensions import Literal

__all__ = ["databrowser"]


@overload
def databrowser(
    *,
    attributes: Literal[False],
    facet: Union[str, list[str]],
) -> dict[Any, dict[str, Any]]:
    ...


@overload
def databrowser(
    *,
    all_facets: Literal[False],
    facet: Literal[None],
) -> Iterator[str]:
    ...


def databrowser(
    *,
    attributes: bool = False,
    all_facets: bool = False,
    facet: Optional[Union[str, list[str]]] = None,
    multiversion: bool = False,
    relevant_only: bool = False,
    batch_size: int = 10,
    count_facet_values: bool = False,
    **search_facets: Union[str, Path, int, list[str]],
) -> Union[dict[Any, dict[Any, Any]], Iterator[str]]:
    """Find data in the system.

    You can either search for files or data facets (variable, model, ...)
    that are available. The query is of the form key=value. <value> might
    use *, ? as wildcards or any regular expression.

    Parameters
    ----------
    **search_facets: Union[str, Path, in, list[str]]
        The facets to be applied in the data search. If not given
        the whole dataset will be queried.
    all_facets: bool, default: False
        Retrieve all facets (attributes & values) instead of the files.
    facet: Union[str, list[str]], default: None
        Retrieve these facets (attributes & values) instead of the files.
    attributes: bool, default: False
        Retrieve all possible attributes for the current search
        instead of the files.
    multiversion: bool, default: False
        Select all versions and not just the latest version (default).
    relevant_only: bool, default: False
        Show only facets that filter more than one result.
    batch_size: int, default: 10
        Size of the search querey.
    count_facet_values: bool, default: False
        Display only this amount for search results.

    Returns
    -------
    Iterator :
        If ``all_facets`` is False and ``facet`` is None an
        iterator with results.
    dict[Any, dict[str, Any] :
        dictionary for facet results, if ``all_facets`` is True or ``facet``
        was given a value (str or list[str])


    Example
    -------

    Seach for files in the system:

    .. execute_code::

        import freva
        files = freva.databrowser(project='obs*', institute='cpc',
                                  time_frequency='??min',
                                  variable='pr')
        print(files)
        print(next(files))
        for file in files:
            print(file)
            break
        facets = freva.databrowser(project='obs*', attributes=True)
        print(list(facets))


    Search for facets in the system:

    .. execute_code::

        import freva
        all_facets = freva.databrowser(project='obs*', all_facets=True)
        print(all_facets)
        spec_facets = freva.databrowser(project='obs*',
                                        facet=["time_frequency", "variable"])
        print(spec_facets)

    Reverse search: retrieving meta data from a knwon file

    .. execute_code::

        import freva
        from pathlib import Path
        file = ".docker/data/observations/grid/CPC/CPC/cmorph/30min/atmos/30min/r1i1p1/v20210618/pr/pr_30min_CPC_cmorph_r1i1p1_201609020000-201609020030.nc"
        res = freva.databrowser(file=file, all_facets=True)
        print(res)

    """
    facets: list[str] = []
    try:
        # If we don't convert a str to a str mypy will complain.
        f = Path(str(search_facets["file"])).expanduser().absolute()
        search_facets["file"] = f'"\{f}"'
    except KeyError:
        pass
    if isinstance(facet, str):
        facet = [facet]
    facet = facet or []
    facets += [f for f in facet if f]
    latest = not multiversion
    if "version" in search_facets and latest:
        # it makes no sense to look for a specific version just among the latest
        # the speedup is marginal and it might not be what the user expects
        logger.warning("Turning latest off when searching for a specific version.")
        latest = False
    logger.debug("Searching dictionary: %s\n", search_facets)
    if (facets or all_facets) and not attributes:
        out = {}
        search_facets["facet.limit"] = search_facets.pop("facet_limit", -1)
        for att, values in SolrFindFiles.facets(
            facets=facets or None, latest_version=latest, **search_facets
        ).items():
            # values come in pairs: (value, count)
            value_count = len(values) // 2
            if relevant_only and value_count < 2:
                continue
            if count_facet_values:
                out[att] = {v: c for v, c in zip(*[iter(values)] * 2)}
            else:
                out[att] = values[::2]
        return out
    if attributes:
        # select all is none defined but this flag was set
        results = SolrFindFiles.facets(
            facets=facets or None, latest_version=latest, **search_facets
        )
        if relevant_only:
            return (k for k in results if len(results[k]) > 2)
        return (k for k in results)
    return SolrFindFiles.search(
        batch_size=batch_size, latest_version=latest, **search_facets
    )



"""
Created on 23.05.2016

@author: Sebastian Illing
"""
import unittest
from datetime import datetime
import pwd
import os


def test_history_model(test_user):
    # test slurm id
    user, hist = test_user
    slurm_id = hist.slurmId()
    assert slurm_id == "44742"

    # test status_names
    status = hist.status_name()
    assert status == "running"

    # test print history
    _str = hist.__str__(compact=True)


def test_result_model(test_user):
    from evaluation_system.model.history.models import Result

    user, hist = test_user
    r = Result.objects.create(
        history_id=hist,
        output_file="/some/paht/to_file.ext",
        preview_file="/path/to/preview.jpg",
        file_type=Result.Filetype.plot,
    )

    assert Result.objects.filter(id=r.id).exists()

    # test get extension
    assert r.fileExtension() == ".ext"


def test_similar_results(dummy_user, test_user):
    from evaluation_system.model.history.models import History, Configuration
    from evaluation_system.tests.mocks.dummy import DummyPlugin

    hists = []
    uid = os.getuid()
    udata = pwd.getpwuid(uid)
    for i in range(10):
        hists += [
            dummy_user.user.getUserDB().storeHistory(
                tool=DummyPlugin(),
                config_dict={
                    "the_number": 42,
                    "number": 12,
                    "something": "else",
                    "other": "value",
                    "input": "/folder",
                    "variable": "pr",
                },
                status=0,
                uid=udata.pw_name,
            )
        ]

    hist = History.objects.create(
        timestamp=datetime.now(),
        status=History.processStatus.running,
        uid=test_user[0],
        configuration="{'the_number': 42, 'number': 12, 'something': 'else', 'other': 'value', 'input': '/folder', 'variable': 'pr'}",
        tool="dummytool",
        slurm_output="/path/to/slurm-44742.out",
    )

    o = Configuration.objects.filter(history_id=hist)
    sim_res = History.find_similar_entries(o)
    assert len(sim_res) == 0



# coding=utf-8
"""
Created on 10.05.2016

@author: Sebastian Illing
"""

import pytest


def test_infer_type(dummy_env):
    from evaluation_system.api.parameters import (
        ParameterType,
        Integer,
        Float,
        Bool,
        String,
        ValidationError,
    )

    assert ParameterType.infer_type(1).__class__ == Integer
    assert ParameterType.infer_type(1.0).__class__ == Float
    assert ParameterType.infer_type("str").__class__ == String
    assert ParameterType.infer_type(True).__class__ == Bool
    assert ParameterType.infer_type(str).__class__ == String
    with pytest.raises(ValueError):
        assert ParameterType.infer_type(type(str))


def test_parameter_type(dummy_env):
    from evaluation_system.api.parameters import String, ValidationError

    with pytest.raises(ValidationError):
        String(name="foo", max_items=0)
    s_type = String(
        name="foo", default="foo;bar", mandatory=True, max_items=2, item_separator=";"
    )
    assert s_type.to_str("foo;bar") == "foo;bar"
    assert s_type.to_str(["foo", "bar"]) == "foo;bar"
    s_type.item_separator = None
    assert s_type.to_str(["foo", "bar"]) == '["foo", "bar"]'
    assert s_type.parse('["foo"]') == ["foo"]
    assert s_type.parse("foo") == ["foo"]
    assert s_type.__str__() == s_type.get_type() == "String"


def test_parsing(dummy_env):
    from evaluation_system.api.parameters import (
        String,
        Integer,
        Float,
        Bool,
        Range,
        SelectField,
        SolrField,
        InputDirectory,
        Unknown,
        ValidationError,
    )

    test_cases = [
        (String(), [("asd", "asd"), (None, "None"), (1, "1"), (True, "True")], []),
        (
            String(regex="^x.*$"),
            [("xasd", "xasd")],
            ["aaaa", "the x is not in, the right place"],
        ),
        (
            String(regex="x+", max_items=3),
            [("xasd", ["xasd"]), ("xas,sxd,somex", ["xas", "sxd", "somex"])],
            ["missing a letter", "ok x, also ok xx, but not ok"],
        ),
        (
            Integer(),
            [("123", 123), ("0", 0), ("-1", -1), (True, 1)],
            ["+-0", "not a number!!", None],
        ),
        (
            Integer(max_items=2, item_separator=","),
            [
                ("123", [123]),
                ("0,4", [0, 4]),
                ([0, 3.1516], [0, 3]),
                ([0.999], [0]),
                (123.8, [123]),
            ],
            ["+-0", "not a number!!", [1, 2, 3]],
        ),
        (
            Float(),
            [
                ("123", 123.0),
                ("0", 0.0),
                ("-1.3", -1.3),
                ("-1e+2", -100.0),
                ("+2E-2", 0.02),
                (False, 0.0),
                ("12.", 12.0),
                (".42", 0.42),
            ],
            ["+-0", "not a number!!", "123,321.2", None],
        ),
        (
            Bool(),
            [
                ("1", True),
                ("0", False),
                ("True", True),
                ("false", False),
                ("no", False),
                ("YES", True),
            ],
            ["maybe", "", None],
        ),
        (
            Range(),
            [
                ("1960:1970", list(range(1960, 1970 + 1))),
                ("1970,1971,1972", [1970, 1971, 1972]),
                ("1980:5:2000", list(range(1980, 2000 + 1, 5))),
                ("1970:1971-1970", [1971]),
                ("1950:10:1980,1985-1960:1970", [1950, 1980, 1985]),
            ],
            ["1:1:1:1", "", None, "ab:c", []],
        ),
        (
            SelectField(
                options={"option1": "value1", "another_option": "Some other value"}
            ),
            [("value1", "option1"), ("Some other value", "another_option")],
            [("bad value", "")],
        ),
        (SolrField(facet="variable"), [("tas", "tas"), ("pr", "pr")], []),
        (InputDirectory(), [("/home/user", "/home/user")], []),
        (Unknown(), [("test", "test")], []),
    ]
    for case_type, positive_cases, negative_cases in test_cases:
        for expected, result in positive_cases:
            parsed_value = case_type.parse(expected)
            if isinstance(parsed_value, list):
                assert type(parsed_value[0]) == case_type.base_type
            else:
                assert type(parsed_value) == case_type.base_type
            assert parsed_value == result
        for unparseable in negative_cases:
            try:
                with pytest.raises(TypeError):
                    case_type.parse(unparseable)
            except:
                try:
                    with pytest.raises(ValueError):
                        case_type.parse(unparseable)
                except:
                    with pytest.raises(ValidationError):
                        case_type.parse(unparseable)


def test_parameters_dictionary(dummy_env):
    from evaluation_system.api.parameters import (
        String,
        ParameterDictionary,
        ValidationError,
    )

    p1 = String(name="a_param1", default="default default 1")
    p2 = String(
        name="a_param2", default="default default 2", max_items=3, item_separator=":"
    )
    assert p2.parse("a:b:C") == ["a", "b", "C"]
    with pytest.raises(ValueError):
        ParameterDictionary(p1, p2, p2)
    p_dict = ParameterDictionary(p1, p2)
    assert "ParameterDictionary(" in p_dict.__str__()
    assert "a_param2<String>:" in p_dict.__str__()
    assert "a_param1<String>:" in p_dict.__str__()
    assert len(p_dict) == 3
    assert p1.name in p_dict and p_dict[p1.name] == p1.default
    assert p2.name in p_dict and p_dict[p2.name] == p2.default
    assert len(p_dict.parameters()) == 3
    with pytest.raises(ValidationError):
        p_dict.get_parameter("a_pram1")
    # TODO: Fix error
    # self.assertEquals(p_dict.get_parameter(p1.name), p1)
    # self.assertEquals(p_dict.get_parameter(p2.name), p2)

    # Check parameters remain in the order we put them
    params = [String(name="extra_scheduler_options", default="")]
    for i in range(1000):
        params.append(String(name=str(i), default=i))
    p_dict = ParameterDictionary(*params)
    assert sorted(p_dict.keys()) == sorted([p.name for p in params])
    assert list(p_dict.values()) == [p.default for p in params]
    assert p_dict.parameters() == params


def test_parse_arguments(dummy_env):
    from evaluation_system.api.parameters import (
        String,
        ParameterDictionary,
        Integer,
        Float,
        Bool,
        String,
        File,
        Date,
        Range,
        ValidationError,
    )

    p_dict = ParameterDictionary(String(name="a"), String(name="b"))
    res = p_dict.parse_arguments("a=1 b=2".split())
    assert res == dict(a="1", b="2")
    p_dict = ParameterDictionary(Integer(name="a"), Integer(name="b"))
    res = p_dict.parse_arguments("a=1 b=2".split())
    assert res == dict(a=1, b=2)
    # more arguments than those expected
    p_dict = ParameterDictionary(Integer(name="a"))
    with pytest.raises(ValidationError):
        p_dict.parse_arguments("a=1 b=2".split())
    p_dict = ParameterDictionary(
        Integer(name="int"),
        Float(name="float"),
        Bool(name="bool"),
        String(name="string"),
        File(name="file", default="/tmp/file1"),
        Date(name="date"),
        Range(name="range"),
    )
    res = p_dict.parse_arguments("int=1 date=1 bool=1".split())
    assert res == dict(int=1, date="1", bool=True)
    res = p_dict.parse_arguments("int=1 date=1 bool=1".split(), use_defaults=True)
    assert res == dict(
        int=1, date="1", bool=True, extra_scheduler_options="", file="/tmp/file1"
    )
    p_dict2 = ParameterDictionary(
        Bool(name="init"),
        Integer(name="num", default=2, item_separator=",", max_items=1),
    )
    assert p_dict2.parse_arguments("init=0") == dict(init=False)
    assert p_dict2.parse_arguments(
        ["num=0", "num=1", "init=0"], check_errors=False
    ) == dict(num=[0, 1], init=False)
    res = p_dict.parse_arguments(
        "int=1 date=1 bool=1 range=1:5".split(),
        use_defaults=True,
        complete_defaults=True,
    )
    assert res == dict(
        int=1,
        date="1",
        bool=True,
        extra_scheduler_options="",
        range=list(range(1, 6)),
        file="/tmp/file1",
        float=None,
        string=None,
    )

    for arg, parsed_val in [
        ("bool=1", True),
        ("bool=true", True),
        ("bool=TRUE", True),
        ("bool=0", False),
        ("bool=false", False),
        ("bool=False", False),
        ("bool=no", False),
        ("bool=NO", False),
        ("bool=YES", True),
        ("bool", True),  # Special case!
        ("float=1.2", 1.2),
        ("float=-1e-1", -0.1),
        ("float=2E2", 200.0),
        ("string=1", "1"),
        ("string=ä", "ä"),
    ]:
        res = p_dict.parse_arguments(arg.split())
        assert res == {
            arg.split("=")[0]: parsed_val
        }, "Error when parsing %s, got %s" % (arg, res)

    # multiple arguments
    p_dict = ParameterDictionary(
        File(name="file", default="/tmp/file1", max_items=2, item_separator=":"),
        Date(name="date", item_separator="/"),
    )
    assert p_dict.parse_arguments(["file=a:b"]) == dict(file=["a", "b"])
    with pytest.raises(ValidationError):
        p_dict.parse_arguments(["file=a:b:c"])
    with pytest.raises(ValidationError):
        p_dict.parse_arguments(["file=a", "file=b", "file=c"])
    # this should still work since max_items defaults to 1
    # and in that case no splitting happens
    assert p_dict.parse_arguments(["date=a/b"]) == dict(date="a/b")
    assert p_dict.parse_arguments(["file=a", "file=b"]) == dict(file=["a", "b"])


def test_complete(dummy_env):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        File,
        Date,
    )

    p_dict = ParameterDictionary(
        Integer(name="int"), File(name="file", default="/tmp/file1"), Date(name="date")
    )
    conf = dict(int=1)
    p_dict._complete(conf)
    assert conf == {"int": 1, "extra_scheduler_options": "", "file": "/tmp/file1"}
    p_dict._complete(conf, add_missing_defaults=True)
    assert conf == {
        "int": 1,
        "extra_scheduler_options": "",
        "date": None,
        "file": "/tmp/file1",
    }

    assert p_dict._complete() == {"extra_scheduler_options": "", "file": "/tmp/file1"}
    assert p_dict._complete(add_missing_defaults=True) == {
        "int": None,
        "date": None,
        "extra_scheduler_options": "",
        "file": "/tmp/file1",
    }

    # assure default value gets parsed (i.e. validated) when creating Parameter
    p = ParameterDictionary(Integer(name="a", default="0"))
    assert p._complete(add_missing_defaults=True) == {
        "a": 0,
        "extra_scheduler_options": "",
    }


def test_defaults(dummy_env):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Bool,
        Integer,
        File,
        Date,
        Range,
        Float,
        ValidationError,
    )

    # All these should cause no exception
    Bool(name="a", default=True)
    Bool(name="a", default="True")
    Bool(name="a", default="0")
    Float(name="a", default=1.2e-2)
    Float(name="a", default="1.2e-2")
    Integer(name="a", default="1232")
    Range(name="a", default="1:5:100")
    ra = Range(name="a", default="1, 2")
    assert ra.to_str([1, 2]) == "[1, 2]"
    p_dict = ParameterDictionary(
        File(
            name="file",
            default="/tmp/file1",
            mandatory=True,
            max_items=2,
            item_separator=":",
        ),
        Date(name="date", item_separator="/"),
    )
    with pytest.raises(ValidationError):
        p_dict.parse_arguments(["date=2"], use_defaults=False)
    assert p_dict.parse_arguments(["date=2"], use_defaults=True) == {
        "date": "2",
        "extra_scheduler_options": "",
        "file": ["/tmp/file1"],
    }


def test_validate_errors(dummy_env):
    from evaluation_system.api.parameters import (
        ParameterDictionary,
        Integer,
        File,
        Float,
    )

    p_dict = ParameterDictionary(
        Integer(name="int", mandatory=True),
        File(name="file", max_items=2, item_separator=":"),
        Float(name="float", mandatory=True, max_items=2),
    )
    assert p_dict.validate_errors(dict(int=1, float=2.0)) == {}
    assert p_dict.validate_errors({"int": None}) == {
        "too_many_items": [],
        "missing": [("int", 1), ("float", 2)],
    }

    assert p_dict.validate_errors({"int": [1, 2, 3, 4, 5]}) == {
        "too_many_items": [("int", 1)],
        "missing": [("float", 2)],
    }
    assert p_dict.validate_errors({"int": [1, 2, 3, 4, 5]}) == {
        "too_many_items": [("int", 1)],
        "missing": [("float", 2)],
    }


def test_help(dummy_env):
    from evaluation_system.api.parameters import ParameterDictionary, Integer, Float

    p_dict = ParameterDictionary(
        Integer(name="answer", help="just some value", default=42, print_format="%sm"),
        Float(
            name="other",
            help="just some super float",
            default=71.7,
            print_format="%.2f",
        ),
    )


def test_special_cases(dummy_env):
    from evaluation_system.api.parameters import Range

    # test __str__ method of Range()
    test_cases = [
        ("1960:1970", range(1960, 1970 + 1)),
        ("1970,1971,1972", [1970, 1971, 1972]),
        ("1980:5:2000", range(1980, 2000 + 1, 5)),
        ("1970:1971-1970", [1971]),
        ("1950:10:1980,1985-1960:1970", [1950, 1980, 1985]),
    ]
    case_type = Range()
    for case, result in test_cases:
        assert str(case_type.parse(case)) == ",".join(map(str, result))


def test_parameter_options(dummy_env):
    from evaluation_system.api.parameters import SelectField, SolrField

    # Arguments of SelectField
    with pytest.raises(TypeError):
        SelectField()
    opts = {"key": "val"}
    p = SelectField(options=opts)
    assert opts == p.options
    # Arguments of SolrField
    with pytest.raises(TypeError):
        SolrField()
    p = SolrField(facet="variable")
    # defaults:
    options = {
        "group": 1,
        "multiple": False,
        "predefined_facets": None,
        "editable": True,
    }
    for key, val in options.items():
        assert getattr(p, key) == val
    options = {
        "facet": "variable",
        "group": 2,
        "multiple": True,
        "predefined_facets": {"key": "val"},
        "editable": False,
    }
    p = SolrField(**options)
    for key, val in options.items():
        assert getattr(p, key) == val



import re
import os


class Esgf2SolrConfig(object):
    def cmip5(self, dataset):
        filename = dataset["title"]
        cmip5_structure = dict(
            filename=filename,
            project=dataset["project"][0],
            product=dataset["product"][0],
            institute=dataset["institute"][0],
            model=re.sub(r"\)", "", re.sub(r"[\.\(\,]", "-", dataset["model"][0])),
            experiment=dataset["experiment"][0],
            time_frequency=dataset["time_frequency"][0],
            realm=dataset["realm"][0],
            cmor_var=[i for i in dataset["variable"] if i + "_" in filename][0],
            ensemble=dataset["ensemble"][0],
        )

        return cmip5_structure

    def specs(self, dataset):
        filename = dataset["title"]
        year = str(
            int(
                re.compile(r"_S(?P<year>\d{4})\d{2}\d{2}_" + dataset["ensemble"][0])
                .search(filename)
                .group("year")
            )
            + 0
        )

        specs_structure = dict(
            filename=re.sub(
                dataset["experiment"][0] + "_S\d{8}_" + dataset["ensemble"][0],
                dataset["experiment"][0] + year + "_" + dataset["ensemble"][0],
                filename,
            ),
            project=dataset["project"][0],
            product=dataset["product"][0],
            institute=dataset["institute"][0],
            model=re.sub(r"\)", "", re.sub(r"[\.\(\,]", "-", dataset["model"][0])),
            experiment=dataset["experiment"][0] + year,
            time_frequency=dataset["time_frequency"][0],
            realm=dataset["realm"][0],
            cmor_var=[i for i in dataset["variable"] if i + "_" in filename][0],
            ensemble=dataset["ensemble"][0],
        )

        return specs_structure

    def project_select(self, dataset):
        options = {
            "CMIP5": self.cmip5,
            "specs": self.specs,
        }
        return options[dataset["project"][0]](dataset)


class Solr2EsgfConfig(object):
    def cmip5(self, esgfpath, filename):

        (
            project,
            product,
            institute,
            model,
            experiment,
            time_frequency,
            realm,
            variable,
            ensemble,
        ) = esgfpath[1:].split("/")

        cmip5_structure = dict(
            project=project,
            product=product,
            institute=institute,
            model=model,
            experiment=experiment,
            time_frequency=time_frequency,
            realm=realm,
            variable=variable,
            ensemble=ensemble,
            filename=filename,
        )
        return cmip5_structure

    def specs(self, esgfpath, filename):

        (
            project,
            product,
            institute,
            model,
            experiment,
            time_frequency,
            realm,
            variable,
            ensemble,
        ) = esgfpath[1:].split("/")

        filename, extension = os.path.os.path.splitext(filename)

        start_end_time = re.sub(r"^.*?" + ensemble, "", filename)[1:]
        specs_experiment = re.sub(r"\d{4}$", "", experiment)
        if start_end_time != "":
            starttime, endtime = start_end_time.split("-")
            if len(starttime) == 6:
                starttime = starttime + "01"
        elif start_end_time == "" and time_frequency == "fx":
            starttime = str(int(re.sub(specs_experiment, "", experiment)) + 1) + "0101"
        filename = (
            re.sub(experiment, "%s_S%s" % (specs_experiment, starttime), filename)
            + extension
        )
        experiment = specs_experiment
        specs_structure = dict(
            project=project,
            product=product,
            institute=institute,
            model=model,
            experiment=experiment,
            time_frequency=time_frequency,
            realm=realm,
            variable=variable,
            ensemble=ensemble,
            filename=filename,
        )
        return specs_structure

    def project_select(self, esgfpath, filename):

        options = {
            "CMIP5": self.cmip5,
            "specs": self.specs,
        }
        return options[esgfpath[1:].split("/")[0]](esgfpath, filename)



import pytest
import mock
import os

SUBCOMMANDS = ("databrowser", "esgf", "crawl-my-data", "history", "plugin")


def test_main_help(dummy_env, capsys):

    from freva.cli import main as main_cli

    with pytest.raises(SystemExit):
        main_cli(["-h"])
    doc_string = capsys.readouterr().out
    for subcommand in SUBCOMMANDS:
        assert subcommand in doc_string


def test_subcommand_help(dummy_env, capsys):
    from freva.cli import main as main_cli
    from freva.cli.databrowser import main as databrowser
    from freva.cli.esgf import main as esgf
    from freva.cli.plugin import main as plugin
    from freva.cli.history import main as history
    from freva.cli.crawl_my_data import main as crawl_my_data

    functions = (databrowser, esgf, crawl_my_data, history, plugin)
    for subcommand, func in dict(zip(SUBCOMMANDS, functions)).items():
        with pytest.raises(SystemExit):
            func(["--help"])
        doc_string1 = " ".join(capsys.readouterr().out.replace(" [-V]", "").split())
        with pytest.raises(SystemExit):
            main_cli([subcommand, "--help"])
        doc_string2 = " ".join(
            capsys.readouterr()
            .out.replace(f"freva {subcommand}", f"freva-{subcommand}")
            .split()
        )
        for line in doc_string2.split("\n"):
            print(line)
            assert line.strip() in doc_string1
        with pytest.raises(SystemExit):
            main_cli([subcommand[:-1]])
        doc_string3 = capsys.readouterr().err
        assert subcommand in doc_string3


def test_admin_subcommans_help(admin_env, capsys):
    from evaluation_system.tests import run_cli

    with mock.patch.dict(os.environ, admin_env, clear=True):
        with pytest.raises(SystemExit):
            run_cli("solr")
        assert "the following sub-commands are required:" in capsys.readouterr().err
        with pytest.raises(SystemExit):
            run_cli("check")
        assert "the following sub-commands are required:" in capsys.readouterr().err



"""Update user data in the apache solr data search server."""

from __future__ import annotations
import logging
from pathlib import Path
from typing import Optional, Union

from evaluation_system.model.user import User
from evaluation_system.misc import config, logger
from evaluation_system.misc.exceptions import ValidationError, ConfigurationException
from evaluation_system.model.solr_core import SolrCore
from evaluation_system.api.user_data import DataReader

__all__ = ["crawl_my_data"]


def _validate_user_dirs(*crawl_dirs: Optional[Union[str, Path]]) -> tuple[Path, ...]:

    try:
        root_path = DataReader.get_output_directory() / f"user-{User().getName()}"
    except ConfigurationException:
        config.reloadConfiguration()
        root_path = DataReader.get_output_directory() / f"user-{User().getName()}"
    user_paths: tuple[Path, ...] = ()
    for crawl_dir in crawl_dirs or (root_path,):
        crawl_dir = Path(crawl_dir or root_path).expanduser().absolute()
        try:
            cr_dir = crawl_dir.relative_to(root_path)
        except ValueError as error:
            raise ValidationError(
                f"You are only allowed to crawl data in {root_path}"
            ) from error
        user_paths += (crawl_dir,)
    return user_paths


def crawl_my_data(*crawl_dirs: Optional[Union[str, Path]], dtype: str = "fs") -> None:
    """Crawl user output data to reingest it into the solr server.

    The data needs to be of a certain strucutre.

    Parameters
    ----------
    crawl_dirs:
        The data path(s) that needs to be crawled
    dtype:
        The data type, currently only files on the file system are supported.

    Raises
    ------
    ValidationError:
        If crawl_dirs do not belong to current user.

    Example
    -------

    .. execute_code::

        import freva
        freva.crawl_my_data()

    """
    if dtype not in ("fs",):
        raise NotImplementedError("Only data on POSIX file system is supported")
    log_level = logger.level
    try:
        logger.setLevel(logging.ERROR)
        print("Status: crawling ...", end="", flush=True)
        for crawl_dir in _validate_user_dirs(*crawl_dirs):
            SolrCore.load_fs(
                crawl_dir,
                chunk_size=1000,
                abort_on_errors=True,
                drs_type=DataReader.drs_specification,
            )
        print("ok", flush=True)
    finally:
        logger.setLevel(log_level)



import sys

from evaluation_system.api.workload_manager.oar import OARJob


def test_header():
    with OARJob(walltime="00:02:00", processes=4, cores=8, memory="28GB") as cluster:
        assert "#OAR -n worker" in cluster.job_header
        assert "#OAR -l /nodes=1/core=8,walltime=00:02:00" in cluster.job_header
        assert "#OAR --project" not in cluster.job_header
        assert "#OAR -q" not in cluster.job_header

    with OARJob(
        queue="regular",
        project="DaskOnOar",
        processes=4,
        cores=8,
        memory="28GB",
        job_extra=["-t besteffort"],
    ) as cluster:
        assert "walltime=" in cluster.job_header
        assert "#OAR --project DaskOnOar" in cluster.job_header
        assert "#OAR -q regular" in cluster.job_header
        assert "#OAR -t besteffort" in cluster.job_header

    with OARJob(cores=4, memory="8GB") as cluster:
        assert "#OAR -n worker" in cluster.job_header
        assert "walltime=" in cluster.job_header
        assert "#OAR --project" not in cluster.job_header
        assert "#OAR -q" not in cluster.job_header


def test_job_script():
    with OARJob(walltime="00:02:00", processes=4, cores=8, memory="28GB") as cluster:
        job_script = cluster.job_script()
        assert "#OAR" in job_script
        assert "#OAR -n worker" in job_script
        assert "#OAR -l /nodes=1/core=8,walltime=00:02:00" in job_script
        assert "#OAR --project" not in job_script
        assert "#OAR -q" not in job_script

        assert "export " not in job_script

    with OARJob(
        walltime="00:02:00",
        processes=4,
        cores=8,
        memory="28GB",
        env_extra=[
            'export LANG="en_US.utf8"',
            'export LANGUAGE="en_US.utf8"',
            'export LC_ALL="en_US.utf8"',
        ],
    ) as cluster:
        job_script = cluster.job_script()
        assert "#OAR" in job_script
        assert "#OAR -n worker" in job_script
        assert "#OAR -l /nodes=1/core=8,walltime=00:02:00" in job_script
        assert "#OAR --project" not in job_script
        assert "#OAR -q" not in job_script

        assert 'export LANG="en_US.utf8"' in job_script
        assert 'export LANGUAGE="en_US.utf8"' in job_script
        assert 'export LC_ALL="en_US.utf8"' in job_script



"""
Created on 27.04.2012

@author: Robert Schuster

Some useful methods to work with shell scripts
"""

import subprocess
import os
from Logger import Logger
import tempfile
import shutil
import atexit
import multiprocessing


class TempDir(object):
    """
    This class represents a temporal directory which is automatically deleted on exit
    """

    def __init__(self, parentdir=None):
        """
        @param parentdir: the directory is created in the parent directory. If the argument
        is none, then the directory is created in the default temp-Directory
        """
        # create the directory
        self.__path = tempfile.mkdtemp(dir=parentdir)

        # register the cleanup-method
        atexit.register(self.cleanup)

    def cleanup(self):
        """
        Remove the temporal directory and its content.
        This method is automatically called on exit
        """
        # check is the directory is still there
        if not os.path.exists(self.__path):
            return

        # delete the directory
        shutil.rmtree(self.__path)

    def getPath(self):
        """
        get the path of the temporal directory
        """
        return self.__path


class ShellScript(object):
    """
    This Class represents a shell script or any other external program
    """

    # the default environment variables
    default_env = os.environ

    @classmethod
    def addDefaultEnvironmentVariable(cls, varname, varvalue):
        """
        Add Environment Variables that are forwarded to every process
        """
        cls.default_env[varname] = varvalue

    @classmethod
    def getstatusoutput(cls, cmd, workpath=None):
        """Return (status, output, cmd) of executing cmd in a shell."""
        """This new implementation should work on all platforms."""
        pipe = subprocess.Popen(
            cmd,
            shell=True,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=cls.default_env,
            cwd=workpath,
        )
        try:
            output = str.join("", pipe.stdout.readlines())
            sts = pipe.wait()
        except KeyboardInterrupt:
            Logger.Error("process killed by KeyboardInterrupt!")
            pipe.kill()
        if sts is None:
            sts = 0
        return sts, output, cmd

    @classmethod
    def run_scripts_parallel(cls, scripts, nproc=None):
        """
        Run multiple scripts in parallel
        @param scrpits  list object containing ShellScript objects
        @param nproc    number of parallel processes, None=Number of available cpus
        """
        if nproc is None:
            nproc = min(multiprocessing.cpu_count(), len(scripts))
        # create the multiprocessing pool
        pool = multiprocessing.Pool(nproc)
        exitcodes = []
        try:
            res = pool.map_async(_run_script, scripts, callback=exitcodes.append)
            res.get()
        except KeyboardInterrupt:
            Logger.Error("Execution canceled by user!", -1)
            pool.terminate()
        # close the pool again and wait for all processes
        pool.close()
        pool.join()
        return exitcodes[0]

    def checkout(self, output, err_text, err_code, print_without_error=True):
        """
        check the output of getstatusoutput for errors.
        """
        if print_without_error:
            print(output[1])
        if output[0] != 0:
            Logger.Error(err_text)
            Logger.Error(self.getCommand(), errorcode=err_code)

    def __init__(self, scriptfile, workpath=None, check=True):
        """
        create the ShellScript object and check the existence of the script file
        """
        self.scriptfile = scriptfile
        self.arguments = []
        self.workpath = workpath

        # check the existence of the script file
        if (
            check
            and workpath is None
            and not os.path.exists(self.scriptfile)
            or workpath is not None
            and not os.path.exists(workpath + "/" + self.scriptfile)
        ):
            Logger.Error("Script " + self.scriptfile + " not found!")
            exit()

    def addPositionalArgument(self, argument):
        """
        add a positional argument to the script
        """
        self.arguments.append((None, argument))

    def addFlag(self, flag, value=None):
        """
        add a non positional argument with a flag to the program
        """
        self.arguments.append((flag, value))

    def __getArgValueString(self, argument):
        """
        @return: A string representation of the argument
        """
        result = ""
        # call the method recursive if a list is found
        if isinstance(argument, list):
            for x in range(len(argument)):
                result += self.__getArgValueString(argument[x])
                if x < len(argument) - 1:
                    result += " "
        # string type
        elif isinstance(argument, str):
            if (
                " " in argument
                or "?" in argument
                or "(" in argument
                or ")" in argument
                or "'" in argument
            ):
                result += "'" + argument + "'"
            else:
                result += argument
        # integer of float
        elif isinstance(argument, int) or isinstance(argument, float):
            result += str(argument)
        # temporal directory
        elif isinstance(argument, TempDir):
            result += "'" + argument.getPath() + "'"
        elif argument is None:
            pass
        # unsupported argument
        else:
            Logger.Error(
                "ShellScripthelper, __getArgValueString: unsupported argument type "
                + str(type(argument))
            )
        return result

    def run(self):
        """
        Run the Script.
        @return: a tuple with the return value
        """

        # run the script
        return self.getstatusoutput(self.getCommand(), self.workpath)

    def getCommand(self):
        """
        returns the command together with its arguments
        """
        # create a list with arguments
        args = ""
        # loop over all arguments
        for argn in self.arguments:
            if argn[0] is None:
                args += " " + self.__getArgValueString(argn[1])
            else:
                args += " " + argn[0] + " " + self.__getArgValueString(argn[1])

        return self.scriptfile + args


# a module method is needed for parallel execution
def _run_script(script):
    return script.run()



"""
Created on 18.05.2016

@author: Sebastian Illing
"""
import os
import datetime
import pwd
import sys
import pytest


from evaluation_system.tests import run_cli
from evaluation_system.tests.mocks.dummy import DummyPlugin


def test_freva_history_method(dummy_history, dummy_user):
    from freva import history

    config_dict = {
        "the_number": 42,
        "number": 12,
        "something": "else",
        "other": "value",
        "input": "/folder",
        "variable": "pr",
    }
    hist_ids = []
    uid = os.getuid()
    udata = pwd.getpwuid(uid)
    for i in range(10):
        hist_ids += [
            dummy_user.user.getUserDB().storeHistory(
                tool=DummyPlugin(), config_dict=config_dict, status=0, uid=udata.pw_name
            )
        ]
    hist = history()
    assert isinstance(hist, (list, tuple))
    assert len(hist) == 10
    for h in hist:
        assert isinstance(h, dict)
        assert h["configuration"] == config_dict
    for method in (str, int, float):
        hist = history(entry_ids=method(hist_ids[0]))
        assert len(hist) == 1
    with pytest.raises(ValueError):
        hist = history(entry_ids="bla")
    hist = history(entry_ids="0")
    assert len(hist) == 0


def test_history_cmd(capsys, dummy_history, dummy_user):

    from freva.cli.history import main as run

    hist_ids = []
    uid = os.getuid()
    udata = pwd.getpwuid(uid)
    for i in range(10):
        hist_ids += [
            dummy_user.user.getUserDB().storeHistory(
                tool=DummyPlugin(),
                config_dict={
                    "the_number": 42,
                    "number": 12,
                    "something": "else",
                    "other": "value",
                    "input": "/folder",
                    "variable": "pr",
                },
                status=0,
                uid=udata.pw_name,
            )
        ]
    # test history output
    run(["-d"])
    output_str = capsys.readouterr().out
    assert output_str.count("dummyplugin") == 10
    assert output_str.count("\n") == 10
    # test limit output
    run_cli(["history", "--limit=3"])
    output_str = capsys.readouterr().out
    assert output_str.count("dummyplugin") == 3
    assert output_str.count("\n") == 3
    # test no result
    output_str = capsys.readouterr().out
    run_cli(["history", "--plugin", "blabla"])
    output_str = capsys.readouterr().out
    assert len(output_str) < 2
    # test return_command option
    run_cli(["history", f"--entry-ids={hist_ids[0]}", "--return-command"])
    output_str = capsys.readouterr().out.strip("\n").strip()
    check_string = [
        "plugin",
        "dummyplugin",
        "something=else",
        "input=/folder",
        "other=value",
        "number=12",
        "the_number=42",
        "variable=pr",
    ]
    for string in check_string:
        assert string in output_str



from evaluation_system.api.workload_manager.local import LocalJob
import pytest


def test_header():
    with LocalJob(walltime="00:02:00", processes=4, cores=8) as cluster:
        cluster.start()
        assert cluster.job_header == ""


def test_job_script():
    with LocalJob("test", env_extra=["source test"], log_directory="/tmp") as cluster:
        cluster.start()
        job_script = cluster.job_script()

        assert "source test" in job_script
        assert "/tmp/test" in job_script

    with LocalJob(
        env_extra=[
            'export LANG="en_US.utf8"',
            'export LANGUAGE="en_US.utf8"',
            'export LC_ALL="en_US.utf8"',
        ],
        log_directory="/tmp",
    ) as cluster:
        cluster.start()
        job_script = cluster.job_script()
        assert 'export LANG="en_US.utf8"' in job_script
        assert 'export LANGUAGE="en_US.utf8"' in job_script
        assert 'export LC_ALL="en_US.utf8"' in job_script
        assert "/tmp/worker" in job_script



"""Deploy the evaluation_system / core."""
import argparse
from configparser import ConfigParser, ExtendedInterpolation
import logging
import hashlib
import os
from os import path as osp
from pathlib import Path
import shlex
import sys
from subprocess import CalledProcessError, PIPE, run
import urllib.request
from tempfile import TemporaryDirectory

DEFAULT_PYTHON = "3.10"
MINICONDA_URL = "https://repo.anaconda.com/miniconda/"
ANACONDA_URL = "https://repo.anaconda.com/archive/"
CONDA_PREFIX = os.environ.get("CONDA", "Anaconda3-2021.11")
CONDA_VERSION = "{conda_prefix}-{arch}.sh"

logging.basicConfig(format="%(name)s - %(levelname)s - %(message)s", level=logging.INFO)
logger = logging.getLogger(__file__)

MODULE = """#%Module4.0 #######################################################
##
## FREVA - Free Evaluation System Framework modulefile
##
#
### BEGIN of config part ********
#define some variables
set shell [module-info shelltype]
set curMode [module-info mode]
module-whatis   "evaluation_system {version}"
proc ModulesHelp {{ }} {{
    puts stderr "Load the free evaluation system framework {project}"
}}
if {{ $curMode eq "load" }} {{
    if {{ $shell == "fish" }} {{
        puts "source {eval_conf_file.parent}/completions/complete_fish"
    }} elseif {{ $shell == "csh" }} {{
        puts "source {eval_conf_file.parent}/completions/complete_csh"
    }} elseif {{ $shell == "sh" }} {{
        puts ". {eval_conf_file.parent}/completions/complete_sh"
    }}
}}
prepend-path PATH {root_dir}/bin
setenv EVALUATION_SYSTEM_CONFIG_FILE {eval_conf_dir}/evaluation_system.conf
setenv EVALUATION_SYSTEM_CONFIG_DIR {eval_conf_dir}
"""

FISH_SCRIPT = """set -g EVALUATION_SYSTEM_CONFIG_FILE {eval_conf_dir}/evaluation_system.conf
set -g EVALUATION_SYSTEM_CONFIG_DIR {eval_conf_dir}
set -gx PATH {root_dir}/bin $PATH
{completion}
"""
FISH_COMPLETION = """source {root_dir}/share/fish/completions/freva.fish
"""

SH_SCRIPT = """export EVALUATION_SYSTEM_CONFIG_FILE={eval_conf_dir}/evaluation_system.config
export EVALUATION_SYSTEM_CONFIG_DIR={eval_conf_dir}
export PATH={root_dir}/bin:$PATH
shell=$(basename $SHELL)
{completion}
"""


SH_COMPLETION = """if [ -z "$shell" ];then
    shell=$(basename $SHELL)
fi
if [ $shell = zsh ];then
    source {root_dir}/share/zsh/site-functions/source.zsh
elif [ $shell = bash ];then
    source {root_dir}/share/bash-completion/completions/freva
fi
"""

CSH_SCRIPT = """setenv PATH {root_dir}/bin\:$PATH
setenv EVALUATION_SYSTEM_CONFIG_FILE "{eval_conf_dir}/evaluation_system.conf"
setenv EVALUATION_SYSTEM_CONFIG_DIR "{eval_conf_dir}"
{completion}
"""


CSH_COMPLETION = """if ( `basename $SHELL` == tcsh ) then
    source {root_dir}/share/tcsh-completion/completion/freva
endif
"""


def get_data_dirs(install_prefix, user):

    install_prefix = Path(install_prefix or sys.prefix)
    if user:
        import appdirs

        data_dir = Path(appdirs.user_data_dir()).parent
        root_dir = Path(appdirs.user_data_dir())
        config_dir = Path(appdirs.user_config_dir()) / "freva"
    else:
        root_dir = data_dir = Path(install_prefix)
        config_dir = root_dir / "freva"
    config_dir.mkdir(exist_ok=True, parents=True)
    return config_dir, data_dir, root_dir


def get_script_path():
    return osp.dirname(osp.realpath(sys.argv[0]))


def read(*parts):
    return open(osp.join(get_script_path(), *parts)).read()


def find_files(path, glob_pattern="*"):
    return [str(f) for f in Path(path).rglob(glob_pattern)]


def find_version(*parts):

    vers_file = read(*parts)
    old_path = sys.path.copy()
    with TemporaryDirectory() as td:
        with open(osp.join(td, "tmp_frevaversion.py"), "w") as f:
            f.write(vers_file)
        sys.path.insert(0, td)
        try:
            from tmp_frevaversion import __version__

            sys.path = old_path
            return __version__
        except ImportError:
            sys.path = old_path
    raise RuntimeError("Unable to find version string.")


def reporthook(count, block_size, total_size):
    if count == 0:
        return
    frac = count * block_size / total_size
    percent = int(100 * frac)
    bar = "#" * int(frac * 40)
    msg = "Downloading: [{0:<{1}}] | {2}% Completed".format(bar, 40, round(percent, 2))
    print(msg, end="\r", flush=True)
    if frac >= 1:
        print()


def parse_args(argv=None):
    """Consturct command line argument parser."""

    ap = argparse.ArgumentParser(
        prog="deploy_freva",
        description="""This Programm installs the evaluation_system package.""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    ap.add_argument(
        "install_prefix", type=Path, help="Install prefix for the environment."
    )
    ap.add_argument(
        "--packages",
        type=str,
        nargs="*",
        help="Pacakges that are installed",
        default=[],
    )
    ap.add_argument(
        "--channel", type=str, default="conda-forge", help="Conda channel to be used"
    )
    ap.add_argument("--shell", type=str, default="bash", help="Shell type")
    ap.add_argument(
        "--arch",
        type=str,
        default="Linux-x86_64",
        choices=[
            "Linux-aarch64",
            "Linux-ppc64le",
            "Linux-s390x",
            "Linux-x86_64",
            "MacOSX-x86_64",
        ],
        help="Choose the architecture according to the system",
    )
    ap.add_argument(
        "--no_conda",
        "--no-conda",
        action="store_true",
        default=False,
        help="Do not create conda environment",
    )
    ap.add_argument(
        "--run_tests",
        action="store_true",
        default=False,
        help="Run unittests after installation",
    )
    ap.add_argument(
        "--editable",
        "-e",
        action="store_true",
        default=False,
        help="Apply pip install in editable mode.",
    )

    ap.add_argument(
        "--silent",
        "-s",
        action="store_true",
        default=False,
        help="Minimize writing to stdout",
    )
    args = ap.parse_args()
    return args


class Installer:
    @property
    def conda_name(self):

        return self.install_prefix.name

    def run_cmd(self, cmd, **kwargs):
        """Run a given command."""
        verbose = kwargs.pop("verbose", False)
        kwargs["check"] = False
        if self.silent and not verbose:
            kwargs["stdout"] = PIPE
            kwargs["stderr"] = PIPE
        res = run(shlex.split(cmd), **kwargs)
        if res.returncode != 0:
            try:
                print(res.stderr.decode())
            except AttributeError:
                # stderr wasn't piped
                pass
            raise CalledProcessError(res.returncode, cmd)

    def use_or_download_temp_conda(self, tempdir):
        """Return to path an existing conda env, if there is none, cerate one."""

        conda_exec_path = Path(os.environ.get("CONDA_EXEC_PATH", ""))
        if conda_exec_path.exists() and conda_exec_path.is_file():
            return Path(conda_exec_path)
        tmp_env = Path(tempdir) / "env"
        conda_script = Path(tempdir) / "anaconda.sh"
        logger.info(f"Downloading {CONDA_PREFIX} script")
        kwargs = {"filename": str(conda_script)}
        if self.silent is False:
            kwargs["reporthook"] = reporthook
        urllib.request.urlretrieve(
            self.conda_url
            + CONDA_VERSION.format(arch=self.arch, conda_prefix=CONDA_PREFIX),
            **kwargs,
        )
        self.check_hash(conda_script)
        conda_script.touch(0o755)
        cmd = f"{self.shell} {conda_script} -p {tmp_env} -b -f"
        logger.info(f"Installing {CONDA_PREFIX}:\n{cmd}")
        self.run_cmd(cmd)
        return tmp_env / "bin" / "conda"

    def create_conda(self):
        """Create the conda environment."""
        with TemporaryDirectory(prefix="conda") as td:
            conda_exec_path = self.use_or_download_temp_conda(td)
            cmd = f"{conda_exec_path} {self.create_command(td)}"
            logger.info(f"Creating conda environment:\n{cmd}")
            self.run_cmd(cmd)

    def create_command(self, tmp_dir):
        """Construct the conda create command."""
        # If packages were given, create a conda env from this packages list
        if self.packages:
            packages = set(self.packages + ["conda", "pip"])
            return (
                f"create -c {self.channel} -q -p {self.install_prefix} "
                f"python={self.python} " + " ".join(packages) + " -y"
            )
        # This is awkward, but since we can't guarrantee that we have a yml
        # parser installed we have to do this manually
        env_file = Path(__file__).parent / "dev-environment.yml"
        return f"env create -q -p {self.install_prefix} -f {env_file} --force"

    def check_hash(self, filename):
        archive = urllib.request.urlopen(self.conda_url).read().decode()
        md5sum = ""
        for line in archive.split("</tr>"):
            if CONDA_VERSION.format(arch=self.arch, conda_prefix=CONDA_PREFIX) in line:
                md5sum = line.split("<td>")[-1].strip().strip("</td>")
        md5_hash = hashlib.md5()
        with filename.open("rb") as f:
            for byte_block in iter(lambda: f.read(4096), b""):
                md5_hash.update(byte_block)
        if md5_hash.hexdigest() != md5sum:
            raise ValueError("Download failed, md5sum mismatch: {md5sum} ")

    def pip_install(self, editable=False):
        """Install additional packages using pip."""

        if editable:
            cmd = f"{self.python_prefix} -m pip install -e .[test]"
        else:
            cmd = f"{self.python_prefix} -m pip install .[test]"
        logger.info(f"Installing additional packages\n{cmd}")
        self.run_cmd(cmd)

    def __init__(
        self,
        install_prefix,
        no_conda=False,
        packages=[],
        channel="conda-forge",
        shell="bash",
        arch="Linux-x86_64",
        python="3.10",
        run_tests=False,
        silent=False,
    ):
        self.run_tests = run_tests
        self.install_prefix: Path = Path(install_prefix).expanduser().absolute()
        self.packages = packages
        self.channel = channel
        self.arch = arch
        self.python = python
        self.silent = silent
        self.shell = shell
        if self.silent:
            logger.setLevel(logging.ERROR)
        self.conda_url = ANACONDA_URL
        if "miniconda" in CONDA_PREFIX.lower():
            self.conda_url = MINICONDA_URL
        self.conda = no_conda is False

    @staticmethod
    def create_paths_for_config(install_prefix, user):
        """Get and prepare the evaluation_system config file."""

        config_parser = ConfigParser(interpolation=ExtendedInterpolation())
        asset_conf_file = Path(__file__).parent / "assets" / "evaluation_system.conf"
        eval_conf_file = Path(
            os.environ.get("EVALUATION_SYSTEM_CONFIG_FILE", asset_conf_file)
        )
        with eval_conf_file.open("r") as fp:
            config_parser.read_file(fp)
        config_dir, data_dir, root_dir = get_data_dirs(install_prefix, user)
        if not config_parser["evaluation_system"]["root_dir"]:
            config_parser.set("evaluation_system", "root_dir", str(root_dir))
        if eval_conf_file == asset_conf_file:
            with asset_conf_file.open() as fobj:
                tmp_cfg = []
                for line in fobj.readlines():
                    if line.strip().startswith("root_dir"):
                        tmp_cfg.append(f"root_dir={root_dir}\n")
                    else:
                        tmp_cfg.append(line)
            eval_conf_file = config_dir / "evaluation_system.conf"
            if not eval_conf_file.exists():
                with open(eval_conf_file, "w") as fobj:
                    fobj.write("".join(tmp_cfg))
        for key in (
            "preview_path",
            "project_data",
            "base_dir_location",
            "scheduler_output_dir",
        ):
            path = Path(config_parser["evaluation_system"][key])
            if path:
                try:
                    path.mkdir(exist_ok=True, parents=True)
                except PermissionError:
                    logger.warning(f"Could not create path: {path}")
        return eval_conf_file, data_dir

    @staticmethod
    def create_loadscript(install_prefix, user=False):
        """Create the load-script for this installation."""
        config_parser = ConfigParser(interpolation=ExtendedInterpolation())
        install_prefix = Path(install_prefix or sys.prefix)
        eval_conf_file, root_dir = Installer.create_paths_for_config(
            install_prefix, user=user
        )
        with eval_conf_file.open("r") as fp:
            config_parser.read_file(fp)
        shell_scripts = dict(fish=FISH_SCRIPT, csh=CSH_SCRIPT, sh=SH_SCRIPT)
        completions = dict(fish=FISH_COMPLETION, csh=CSH_COMPLETION, sh=SH_COMPLETION)
        for shell in ("fish", "csh", "sh"):
            activate_file = eval_conf_file.parent / f"activate_{shell}"
            source_file = eval_conf_file.parent / "completions" / f"complete_{shell}"
            try:
                source_file.parent.mkdir(parents=True, exist_ok=True)
                with (activate_file).open("w") as f:
                    f.write(
                        shell_scripts[shell].format(
                            root_dir=install_prefix,
                            eval_conf_dir=eval_conf_file.parent,
                            completion=completions[shell].format(root_dir=root_dir),
                        )
                    )
                with (source_file).open("w") as f:
                    f.write(completions[shell].format(root_dir=install_prefix))
            except Exception as error:
                logging.warning("Could not create loadscripts: %s", str(error))
        try:
            with (eval_conf_file.parent / "loadfreva.modules").open("w") as f:
                f.write(
                    MODULE.format(
                        version=find_version("src/evaluation_system", "__init__.py"),
                        root_dir=install_prefix,
                        eval_conf_dir=eval_conf_file.parent,
                        project=config_parser["evaluation_system"]["project_name"],
                    )
                )
        except Exception as error:
            logging.warning("Could not create modulefile: %s", str(error))

    def unittests(self):
        """Run unittests."""
        logger.info("Running unittests:")
        env = os.environ.copy()
        env["PATH"] = f'{self.install_prefix / "bin"}:{env["PATH"]}'
        env["FREVA_ENV"] = str(self.install_prefix / "bin")
        self.run_cmd("make test", verbose=True, env=env)

    @property
    def python_prefix(self):
        """Get the path of the new conda evnironment."""
        return self.install_prefix / "bin" / "python3"


if __name__ == "__main__":
    args = parse_args(sys.argv)
    Inst = Installer(
        install_prefix=args.install_prefix,
        no_conda=args.no_conda,
        packages=args.packages,
        channel=args.channel,
        shell=args.shell,
        arch=args.arch,
        python=os.environ.get("PYTHON_VERSION", DEFAULT_PYTHON),
        run_tests=args.run_tests,
        silent=args.silent,
    )
    if Inst.conda:
        Inst.create_conda()
        Inst.pip_install(args.editable)
    Inst.create_loadscript(Inst.install_prefix)
    if Inst.run_tests:
        Inst.unittests()



"""
Created on 18.05.2016

@author: Sebastian Illing
"""
import logging
from functools import partial
import os
import mock
from pathlib import Path
import pytest
import mock
import time
from subprocess import Popen
from evaluation_system.tests import run_cli, similar_string
from evaluation_system.misc.exceptions import PluginNotFoundError, ValidationError


@mock.patch("os.getpid", lambda: 12345)
def test_cli(dummy_plugin, capsys, dummy_config, caplog):

    from freva.cli.plugin import main as plugin_cli
    import time
    from evaluation_system.misc.exceptions import ValidationError
    from evaluation_system.misc import config

    with pytest.raises(ValidationError):
        plugin_cli(["dummyplugin"])
    plugin_cli(["dummyplugin", "the_number=13"])
    output = capsys.readouterr().out
    assert "the_number" in output
    assert "13" in output
    assert os.getpid() == 12345
    out_path = (
        Path(dummy_config.get("scheduler_output_dir"))
        / "dummyplugin"
        / "DummyPlugin-12345.local"
    )
    try:
        out_path.unlink()
    except FileNotFoundError:
        pass
    plugin_cli(["dummyplugin", "the_number=13"])
    interactive = capsys.readouterr().out
    assert dummy_plugin.plugin_output_file == out_path
    assert out_path.exists()
    with out_path.open() as f:
        assert interactive == f.read()
    out_path.unlink()
    plugin_cli(["dummyplugin", "the_number=13", "--batchmode"])
    output = capsys.readouterr().out
    _, loglevel, message = caplog.record_tuples[-1]
    assert loglevel == logging.INFO
    assert "tail -f" in message
    out_f = Path(message.split("\n")[-1].split(" ")[-1])
    assert out_f.exists()
    with out_f.open() as f:
        assert "pending" in f.read()


def test_killed_jobs_set_to_broken():
    from freva.cli.plugin import main as plugin_cli
    import freva

    cmd = ["freva-plugin", "dummyplugin", "the_number=13", "other=-10"]
    res = Popen(cmd)
    time.sleep(3)
    os.kill(res.pid, 15)
    time.sleep(2)
    hist = freva.history()[0]
    assert hist["status_dict"][hist["status"]].lower() == "broken"


@mock.patch("os.getpid", lambda: 12345)
def test_tool_doc(capsys, plugin_doc, admin_env, caplog):
    cmd = ["doc", "DummyPlugin", "--file-name", str(plugin_doc)]
    with mock.patch.dict(os.environ, admin_env, clear=True):
        run_cli(cmd)
        out = capsys.readouterr().out
        _, loglevel, message = caplog.record_tuples[-1]
        assert loglevel == logging.INFO
        assert "dummyplugin" in message.lower()
        assert "created" in message.lower()
        with pytest.raises(FileNotFoundError):
            run_cli(cmd[:-2])


@mock.patch("os.getpid", lambda: 12345)
def test_forbidden_tool_doc(dummy_env):
    from freva.cli.admin import update_tool_doc

    with pytest.raises(RuntimeError):
        update_tool_doc("dummyplugin")
    with pytest.raises(SystemExit):
        run_cli(["solr", "doc" "--help"])


@mock.patch("os.getpid", lambda: 12345)
def test_list_tools(capsys, dummy_env):
    from freva.cli.plugin import main as run

    run(["--list-tools"])
    plugin_list = capsys.readouterr().out
    assert "DummyPlugin" in plugin_list
    run_cli(["plugin", "dummyplugin", "--doc"])
    help_str = capsys.readouterr().out
    assert similar_string(
        help_str,
        """DummyPlugin (v0.0.0): A dummy plugin
Options:
number                  (default: <undefined>)
                        This is just a number, not really important
the_number              (default: <undefined>) [mandatory]
                        This is *THE* number. Please provide it
something               (default: test)
                        No help available.
other                   (default: 1.4)
                        No help available.
input                   (default: <undefined>)
                        An input file
variable                (default: tas)
                        An input variable
extra_scheduler_options (default: --qos=test, --array=20)
                        Set additional options for the job submission to the
                        workload manager (, seperated). Note: batchmode and web
                        only.""",
    )


@mock.patch("os.getpid", lambda: 12345)
def test_run_pyclientplugin(dummy_history):
    import freva
    from evaluation_system.misc import config
    from evaluation_system.model.plugins.models import ToolPullRequest

    res, _ = freva.run_plugin("dummyplugin", the_number=32, caption="Some caption")
    assert res == 0
    _, res = freva.run_plugin("dummyplugin", the_number=32, show_config=True)
    res = "\n".join([l.strip() for l in res.split("\n") if l.strip()])
    assert similar_string(
        res,
        """    number: -the_number: 32 something: test other: 1.4 input: -variable: tas
extra_scheduler_options: - (default: )""",
    )
    return_val, repo = freva.run_plugin("dummyplugin", repo_version=True)
    assert "repository" in repo.lower()
    assert "version" in repo.lower()

    ToolPullRequest.objects.all().delete()
    with pytest.raises(PluginNotFoundError):
        freva.run_plugin("dummyplugin0", pull_request=True, tag="")

    ret, _ = freva.run_plugin("dummyplugin", pull_request=True, tag="")
    assert ret != 0

    def pr_sleep(t, version=None, status=None, tool="dummyplugin"):
        t = ToolPullRequest.objects.get(tool=tool, tagged_version=version)
        t.status = status
        t.save()

    time.sleep = partial(pr_sleep, version="1.0", status="failed", tool="dummyplugin")
    retun_val, cmd_out = freva.run_plugin("dummyplugin", pull_request=True, tag="1.0")
    assert similar_string(
        cmd_out, """The pull request failed.\nPlease contact the admins.""", 0.7
    )
    time.sleep = partial(pr_sleep, version="2.0", status="success", tool="dummyplugin")
    _, cmd_out = freva.run_plugin("dummyplugin", pull_request=True, tag="2.0")
    assert similar_string(
        cmd_out,
        """dummyplugin plugin is now updated in the system. New version:  2.0""",
        0.7,
    )


@mock.patch("os.getpid", lambda: 12345)
def test_run_plugin(capsys, dummy_history, dummy_env):
    from evaluation_system.misc import config

    with pytest.raises(ValidationError):
        run_cli(["plugin", "dummyplugin"])
    # test run tool
    run_cli(["plugin", "dummyplugin", "the_number=32", '--caption="Some caption"'])
    output_str = capsys.readouterr().out
    assert "Dummy tool was run" in output_str
    assert "the_number" in output_str
    # test get version
    run_cli(["plugin", "dummyplugin", "--repo-version"])
    # test batch mode
    run_cli(["plugin", "dummyplugin", "the_number=32", "--batchmode"])
    # test save config
    run_cli(["plugin", "dummyplugin", "the_number=32", "--save", "--debug"])
    fn = (
        Path(config.get(config.BASE_DIR_LOCATION))
        / "config/dummyplugin/dummyplugin.conf"
    )
    assert not fn.is_file()
    # test show config
    run_cli(["plugin", "dummyplugin", "the_number=42", "--show-config"])
    output_str = capsys.readouterr().out
    for line in (
        "number: -",
        "the_number: 42",
        "something: test",
        "other: 1.4",
        "input: -",
    ):
        assert line in output_str


@mock.patch("os.getpid", lambda: 12345)
def test_handle_pull_request(dummy_env, capsys):
    from evaluation_system.model.plugins.models import ToolPullRequest

    ToolPullRequest.objects.all().delete()
    tool = "dummyplugin"
    run_cli(["plugin", tool, "--pull-request"])
    cmd_out = capsys.readouterr().out
    assert similar_string(cmd_out, """'Missing required option "--tag"'""", 0.7)

    def pr_sleep(t, version=None, status=None, tool="dummyplugin"):

        t = ToolPullRequest.objects.get(tool=tool, tagged_version=version)
        t.status = status
        t.save()

    time.sleep = partial(pr_sleep, version="1.0", status="failed", tool=tool)
    run_cli(["plugin", tool, "--pull-request", "--tag=1.0"])
    cmd_out = capsys.readouterr().out
    assert "The pull request failed.\nPlease contact the admins." in cmd_out
    time.sleep = partial(pr_sleep, version="2.0", status="success", tool=tool)
    run_cli(["plugin", tool, "--pull-request", "--tag=2.0"])
    cmd_out = capsys.readouterr().out
    assert "New version: 2.0" in cmd_out



"""
Created on 20.04.2012

@author: Robert Schuster
"""
from Logger import Logger
from ShellScriptHelper import ShellScript


class NCLscript(ShellScript):
    """
    This class represents an ncl script together with all its arguments
    """

    def addArgument(self, argument, value):
        """
        Add an argument to the script
        @param argument: name of the argument
        @param value: value of the argument
        """
        # check the argument
        if not isinstance(argument, str):
            Logger.Error("NCLhelper, addArgument: the argument name must be a string!")
            return
        self.arguments.append((argument, value))

    def __getArgValueString(self, argument):
        """
        @return: A string representation of the argument
        """
        result = ""
        # call the method recursive if a list is found
        if isinstance(argument, list):
            result += "(/"
            for x in range(len(argument)):
                result += self.__getArgValueString(argument[x])
                if x < len(argument) - 1:
                    result += ","
            result += "/)"
        # string type
        elif isinstance(argument, str):
            result += '"' + argument + '"'
        # integer of float
        elif isinstance(argument, int) or isinstance(argument, float):
            result += str(argument)
        # unsupported argument
        else:
            Logger.Error(
                "NCLhelper, __getArgValueString: unsupported argument type "
                + str(type(argument))
            )
        return result

    def getCommand(self):
        """
        returns the command together with its arguments
        """
        # create a list with arguments
        args = ""
        # loop over all arguments
        for argn in self.arguments:
            args += " '"
            args += argn[0] + "=" + self.__getArgValueString(argn[1])
            args += "'"
        return "ncl -Q %s%s" % (self.scriptfile, args)

    def run(self):
        """
        Run the Script.
        @return: a tuple with the return value
        """
        # run the script
        result = self.getstatusoutput(self.getCommand(), self.workpath)
        if result[0] == 0 and "fatal:" in result[1]:
            result = 1, result[1], result[2]
        if "warning" in result[1]:
            print(result[1])
        if "debug" in result[1]:
            print(result[1])
        return result



"""
Created on 17.05.2016

@author: Sebatian Illing
"""

import pytest
import os


def test_struct():
    from evaluation_system.misc.utils import Struct

    ref_dict = {"a": 1, "b": 2}
    s = Struct(**ref_dict)
    assert "a" in s.__dict__
    assert s.validate(1)
    assert s.a == 1
    assert "b" in s.__dict__
    assert s.b == 2
    assert s.validate(2)
    assert s.toDict() == ref_dict
    assert Struct.from_dict(ref_dict) == s
    assert s.__repr__() == "<a:1,b:2>"

    # recursion
    ref_dict = {"a": {"b": {"c": 1}}}
    s = Struct.from_dict(ref_dict, recurse=True)
    assert "a" in s.__dict__
    assert "b" in s.a.__dict__
    assert "c" in s.a.b.__dict__
    assert s.a.b.validate(1)
    assert s.a.b.c == 1
    assert s.__repr__() == "<a:<b:<c:1>>>"


def test_template_dict():
    from evaluation_system.misc.utils import TemplateDict

    t = TemplateDict(a=1, b=2, c="$a")
    res = t.substitute(dict(x="$b$b$b", y="$z", z="$a"), recursive=False)
    assert res == {"x": "222", "y": "$z", "z": "1"}
    res = t.substitute(dict(x="$b$b$b", y="$c", z="$a"), recursive=True)
    assert res == {"x": "222", "y": "1", "z": "1"}
    tmp = {}
    # the maximal amount depends on de order they get resolved, and in
    # dictionaries this order is not
    # even given by anything in particular (e.g. alphabetic).
    # for a transitive substitution (a=b,b=c,c=d,...) the best case is
    # always 1 and the worst is ceil(log_2(n))
    # We have a maximal_iterations of 15 so we can substitute *at
    # least* 2^15=32768 variables
    max_it = 5000
    for l in range(max_it):
        tmp["a_%s" % l] = "$a_%s" % (l + 1)
    tmp["a_%s" % max_it] = "end"
    res = t.substitute(tmp, recursive=True)
    assert all([r == "end" for r in res.values()])
    # check recursions that doesn't work
    tmp["a_%s" % max_it] = "$a_0"
    with pytest.raises(Exception):
        t.substitute(tmp, recursive=True)
    with pytest.raises(Exception):
        t.substitute(dict(x="$y", y="$x"), recursive=True)
    with pytest.raises(Exception):
        t.substitute(dict(x="$y", y="$z", z="$x"), recursive=True)


def test_metadict_creation():
    from evaluation_system.misc.utils import metadict

    m1 = metadict(dict(a=1, b=2, c=[1, 2, 3]))
    m2 = metadict(a=1, b=2, c=[1, 2, 3])
    assert m1 == m2
    m3 = metadict(a=1, b=2, c=[1, 2, 3])
    m3.setMetadata("a", test=1)
    # metadata is just a parallel storage and should not affect the data.
    assert m1 == m3
    # the  'compact_creation' is a special key!
    m4 = metadict(compact_creation=False, a=1, b=2, c=[1, 2, 3])
    assert m1 == m4
    assert not ("compact_creation" in m4)
    # but after creation you should be able to use it
    m4["compact_creation"] = True
    assert not (m1 == m4)
    assert "compact_creation" in m4
    # setting compact creation to True should only affect tuples! Not lists.
    m5 = metadict(compact_creation=True, a=1, b=2, c=[1, 2, 3])
    assert m1 == m5
    # Should fail if compact_creation is set and values are bad formed (
    # i.e. iff tuple then (value, dict)
    with pytest.raises(AttributeError):
        metadict(compact_creation=True, a=(1, 2), b=2, c=[1, 2, 3])
    with pytest.raises(AttributeError):
        metadict(compact_creation=True, a=(1, [2, 3]), b=2, c=[1, 2, 3])
    # Compact creation should produce the same outcome as the normal one
    m6 = metadict(compact_creation=True, a=(1, dict(test=1)), b=2, c=[1, 2, 3])
    assert m1 == m6
    assert m3.getMetadata("a") == m6.getMetadata("a")


def test_metadict_copy():
    from evaluation_system.misc.utils import metadict

    m = metadict(dict(a=1, b=2, c=[1, 2, 3]))
    n = m.copy()
    n["c"][0] = 0
    # check we have a deepcopy of the items
    assert n["c"][0] != m["c"][0]


def test_printable_list():
    from evaluation_system.misc.utils import PrintableList

    a = [2, 3, 4, 5]
    p_list = PrintableList(a)
    assert p_list == a  # list creation
    assert str(p_list) == "2,3,4,5"  # string functionality
    p_list = PrintableList(a, seperator=":")
    assert str(p_list) == "2:3:4:5"  # string functionality


def test_super_make_dirs(temp_dir):
    from evaluation_system.misc.utils import supermakedirs

    path = temp_dir / "should" / "exist" / "now"
    res = supermakedirs(str(path), 0o777)
    assert path.is_dir
    assert os.access(path, os.W_OK)
    assert os.access(path, os.R_OK)


def test_mp_wrap_fn():
    def test_f(x, y):
        return x * y

    from evaluation_system.misc.utils import mp_wrap_fn

    res = mp_wrap_fn([test_f, 3, 2])
    assert res == 6



"""Module to access the esgf data catalogue."""
from __future__ import annotations
from pathlib import Path
from evaluation_system.model.esgf import P2P
from typing import Union, Optional, overload
from typing_extensions import Literal

__all__ = ["esgf"]


@overload
def esgf(datasets: Literal[True]) -> list[tuple[str, str]]:
    ...


@overload
def esgf(datasets: Literal[False], show_facet: str) -> dict[str, list[str]]:
    ...


def esgf(
    datasets: bool = False,
    show_facet: Optional[Union[str, list[str]]] = None,
    download_script: Optional[Union[str, Path]] = None,
    query: Optional[str] = None,
    opendap: bool = False,
    gridftp: bool = False,
    **search_constraints: dict[str, str],
) -> Union[str, list[tuple[str, str]], dict[str, list[str]]]:
    """Find data in the ESGF.

    ::
        import freva
        files = freva.esgf(model='MPI-ESM-LR',
                           experiment='decadal2001',
                           variable='tas'
                )

    Parameters:
    -----------
    datasets:
        List the name of the datasets instead of showing the urls.
    show_facet:
        List all values for the given facet (might be
        defined multiple times). The results show the possible
        values of the selected facet according to the given
        constraints and the number of *datasets* (not files)
        that selecting such value as a constraint will result
        (faceted search)
    opendap:
        List the name of the datasets instead of showing the urls.
    gridftp:
        Show Opendap endpoints instead of the http default
        ones (or skip them if none found)
    download_script:
        Download wget_script for getting the files
        instead of displaying anything (only http)
    query:
        Display results from <list> queried fields
    **search_constraints:
        Search facets to be applied in the data search.

    Returns:
    --------
    Collection of files, facets or attributes

    """

    result_url = []
    show_facet = show_facet or []
    url_type = "http"
    if opendap:
        url_type = "opendap"
    if gridftp:
        url_type = "gridftp"
    # find the files and display them
    p2p = P2P()
    if download_script:
        download_script = Path(download_script)
        download_script.touch(0o755)
        with download_script.open("bw") as f:
            f.write(p2p.get_wget(**search_constraints))
        return str(
            f"Download script successfully saved to {download_script}"
        )  # there's nothing more to be executed after this
    if datasets:
        return sorted(p2p.get_datasets_names(**search_constraints))
    if isinstance(query, str):
        if len(query.split(",")) > 1:
            # we get multiple fields queried, return in a tructured fashion
            return p2p.show(
                p2p.get_datasets(fields=query, **search_constraints), return_str=True
            )
        else:
            # if only one then return directly
            return p2p.get_datasets(fields=query, **search_constraints)
    if show_facet:
        return p2p.get_facets(show_facet, **search_constraints)
    for result in p2p.files(fields="url", **search_constraints):
        for url_encoded in result["url"]:
            url, _, access_type = url_encoded.split("|")
            if access_type.lower().startswith(url_type):
                result_url.append(url)
    return result_url



"""Submit jobs to the slurm workload manager."""
from __future__ import annotations
import logging
import math
from typing import Optional, ClassVar, Union

from .core import Job

logger = logging.getLogger(__name__)


class SLURMJob(Job):
    # Override class variables
    submit_command: ClassVar[str] = "sbatch"
    cancel_command: ClassVar[str] = "scancel"
    config_name: ClassVar[str] = "slurm"

    def __init__(
        self,
        name: Optional[str] = None,
        queue: Optional[str] = None,
        project: Optional[str] = None,
        walltime: str = "",
        job_cpu: Optional[int] = None,
        job_mem: Optional[str] = None,
        job_extra: Optional[list[str]] = None,
        **base_class_kwargs,
    ):
        super().__init__(name=name, **base_class_kwargs)
        job_extra = job_extra or []
        header_lines = []
        # SLURM header build
        if self.job_name is not None:
            header_lines.append("#SBATCH -J %s" % self.job_name)
        if self.log_directory is not None:
            header_lines.append(
                "#SBATCH -o %s/%s-%%J.out"
                % (self.log_directory, self.job_name or "worker")
            )
        if queue is not None:
            header_lines.append("#SBATCH -p %s" % queue)
        if project is not None:
            header_lines.append("#SBATCH -A %s" % project)

        # Init resources, always 1 task,
        # and then number of cpu is processes * threads if not set
        header_lines.append("#SBATCH -n 1")
        header_lines.append(
            "#SBATCH --cpus-per-task=%d" % (job_cpu or self.worker_cores)
        )
        # Memory
        memory = job_mem
        if job_mem is None:
            memory = slurm_format_bytes_ceil(self.worker_memory)
        if memory is not None:
            header_lines.append("#SBATCH --mem=%s" % memory)

        if walltime is not None:
            header_lines.append("#SBATCH -t %s" % walltime)
        header_lines.extend(["#SBATCH %s" % arg for arg in job_extra])
        header_lines.append("#SBATCH --signal=15@10")
        # Declare class attribute that shall be overridden
        self.job_header = "\n".join(header_lines)


def slurm_format_bytes_ceil(n: Union[int, float]) -> str:
    """Format bytes as text.

    SLURM expects KiB, MiB or Gib, but names it KB, MB, GB. SLURM does
    not handle Bytes, only starts at KB.

    >>> slurm_format_bytes_ceil(1)
    '1K'
    >>> slurm_format_bytes_ceil(1234)
    '2K'
    >>> slurm_format_bytes_ceil(12345678)
    '13M'
    >>> slurm_format_bytes_ceil(1234567890)
    '2G'
    >>> slurm_format_bytes_ceil(15000000000)
    '14G'
    """
    if n >= (1024**3):
        return "%dG" % math.ceil(n / (1024**3))
    if n >= (1024**2):
        return "%dM" % math.ceil(n / (1024**2))
    if n >= 1024:
        return "%dK" % math.ceil(n / 1024)
    return "1K"



from __future__ import annotations
import datetime
import os
import pytest
import mock


def test_broken_command(hist_obj, capsys, admin_env):
    from evaluation_system.tests import similar_string, run_cli
    from django.contrib.auth.models import User
    from evaluation_system.model.history.models import History

    History.objects.create(
        status=History.processStatus.finished,
        slurm_output="/some/out.txt",
        timestamp=datetime.datetime.now(),
        uid=User.objects.first(),
    )
    with mock.patch.dict(os.environ, admin_env, clear=True):
        run_cli(["check", "broken-runs"])
        cmd_out = capsys.readouterr().out
    assert f"Setting job {hist_obj.id} to broken\n" in cmd_out
    assert History.objects.get(id=hist_obj.id).status == History.processStatus.broken


def test_forbidden_usage(capsys, dummy_env):
    from evaluation_system.tests import run_cli
    from freva.cli.admin import check4broken_runs

    with pytest.raises(SystemExit):
        run_cli(["check", "--help"])
    assert "invalid choice" in capsys.readouterr().err
    with pytest.raises(RuntimeError):
        check4broken_runs()



"""Collection of admin commands for apache solr requests."""
from __future__ import annotations
import argparse
from pathlib import Path
from typing import Any, Optional

from ..utils import BaseParser, is_admin, subparser_func_type

from evaluation_system.model.solr_core import SolrCore
from evaluation_system.misc import config


__all__ = ["re_index", "del_index"]


CLI = "SolrCli"


def re_index(
    input_dir: Path,
    *,
    abort_on_errors: bool = False,
    chunk_size: int = 200,
    host: Optional[str] = None,
    port: Optional[int] = None,
) -> None:
    """(Re)-Index data files on posix file system on the apache solr server.

    Parameters:
    ----------
        input_dir:
            The input directory that is crawled, can also be single files
        abort_on_errors:
            Do not keep ingesting data if it fails for one file
        chunk_size:
            Size of the chunks tht is ingested to the solr server
        host:
            The server hostname of the apache solr server.
        port:
            The host port number the apache solr server is listinig to.
    """
    is_admin(raise_error=True)
    SolrCore.load_fs(
        Path(input_dir).expanduser().absolute(),
        chunk_size=max(1, int(chunk_size)),
        abort_on_errors=abort_on_errors,
        host=host,
        port=port,
    )


def del_index(
    file_pattern: Path,
    host: Optional[str] = None,
    port: Optional[int] = None,
) -> None:
    """Delete entries from solr server.

    Parameters:
    ----------
    file_pattern:
        Directory name/File name or Regex expression for entries that
        should be deleted
    host:
        The server hostname of the apache solr server.
    port:
        The host port number the apache solr server is listinig to.
    """
    is_admin(raise_error=True)
    SolrCore.delete_entries(file_pattern, host=host, port=port, prefix="file")


class SolrIndex(BaseParser):
    """Parser for indexing solr server data."""

    desc = "(Re)-Index data on the apache solr server."

    def __init__(self, parser: argparse.ArgumentParser):
        """Construct the parser for indexing data."""
        parser.add_argument(
            "input_dir",
            help="The input directory/file that needs to be (re)-indexed.",
            type=Path,
        )
        parser.add_argument(
            "--chunk-size",
            help="Request size that is submitted to solr server",
            default=200,
            type=int,
        )
        parser.add_argument("--abort-on-errors", action="store_true", default=False)
        parser.add_argument(
            "--delete",
            action="store_true",
            default=False,
            help="Delete entries instead of adding them",
        )
        parser.add_argument(
            "--host",
            type=str,
            default=config.get("solr.host"),
            help="Apache solr server hostname",
        )
        parser.add_argument(
            "--port",
            type=str,
            default=config.get("solr.port"),
            help="Host port the solr server is listning to",
        )
        self.parser = parser
        parser.add_argument(
            "--debug",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.logger.setLevel(20)  # Set log level to info
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(
        args: argparse.Namespace, other_args: Optional[list[str]], **kwargs: Any
    ) -> None:
        """Reindex the data."""
        input_dir = kwargs.pop("input_dir")
        if kwargs.pop("delete"):
            return del_index(input_dir, port=kwargs["port"], host=kwargs["host"])
        re_index(input_dir, **kwargs)


class SolrCli(BaseParser):
    """Interface defining parsers for the solr core."""

    desc = "Apache solr server related sub-commands."

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        """Construct the sub arg. parser."""

        sub_commands: dict[str, subparser_func_type] = {"index": self.parse_index}
        super().__init__(sub_commands, parser)
        # This parser doesn't do anything without a sub-commands
        # hence the default function should just print the usage
        self.parser.set_defaults(apply_func=self._usage)

    @staticmethod
    def parse_index(help: str, subparsers: argparse._SubParsersAction) -> SolrIndex:
        sub_parser = subparsers.add_parser(
            "index",
            description=help,
            help=help,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        return SolrIndex(sub_parser)
